#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit testing for the VIIRS enhancement function."""

import unittest

import dask.array as da
import numpy as np
import xarray as xr

from .utils import run_and_check_enhancement


class TestVIIRSEnhancement(unittest.TestCase):
    """Class for testing the VIIRS enhancement function in satpy.enhancements.viirs."""

    def setUp(self):
        """Create test data."""
        data = np.arange(15, 301, 15).reshape(2, 10)
        data = da.from_array(data, chunks=(2, 10))
        self.da = xr.DataArray(data, dims=("y", "x"), attrs={"test": "test"})
        self.palette = {"colors":
                        [[14, [0.0, 0.0, 0.0]],
                         [15, [0.0, 0.0, 0.39215]],
                         [16, [0.76862, 0.63529, 0.44705]],
                         [17, [0.76862, 0.63529, 0.44705]],
                         [18, [0.0, 0.0, 1.0]],
                         [20, [1.0, 1.0, 1.0]],
                         [27, [0.0, 1.0, 1.0]],
                         [30, [0.78431, 0.78431, 0.78431]],
                         [31, [0.39215, 0.39215, 0.39215]],
                         [88, [0.70588, 0.0, 0.90196]],
                         [100, [0.19607, 1.0, 0.39215]],
                         [120, [0.19607, 1.0, 0.39215]],
                         [121, [0.0, 1.0, 0.0]],
                         [130, [0.0, 1.0, 0.0]],
                         [131, [0.78431, 1.0, 0.0]],
                         [140, [0.78431, 1.0, 0.0]],
                         [141, [1.0, 1.0, 0.58823]],
                         [150, [1.0, 1.0, 0.58823]],
                         [151, [1.0, 1.0, 0.0]],
                         [160, [1.0, 1.0, 0.0]],
                         [161, [1.0, 0.78431, 0.0]],
                         [170, [1.0, 0.78431, 0.0]],
                         [171, [1.0, 0.58823, 0.19607]],
                         [180, [1.0, 0.58823, 0.19607]],
                         [181, [1.0, 0.39215, 0.0]],
                         [190, [1.0, 0.39215, 0.0]],
                         [191, [1.0, 0.0, 0.0]],
                         [200, [1.0, 0.0, 0.0]],
                         [201, [0.0, 0.0, 0.0]]],
                        "min_value": 0,
                        "max_value": 201}

    def test_viirs(self):
        """Test VIIRS flood enhancement."""
        from satpy.enhancements.viirs import water_detection
        expected = [[[1, 7, 8, 8, 8, 9, 10, 11, 14, 8],
                     [20, 23, 26, 10, 12, 15, 18, 21, 24, 27]]]
        run_and_check_enhancement(water_detection, self.da, expected,
                                  palettes=self.palette)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reading and calibrating GAC and LAC AVHRR data.

Uses Pygac under the hood. See the `Pygac Documentation`_ for supported data
formats as well as calibration and navigation methods.

.. todo::

    Fine grained calibration
    Radiance output

.. _Pygac Documentation:
    https://pygac.readthedocs.io/en/stable
"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import pygac.utils
import xarray as xr
from pygac.gac_klm import GACKLMReader
from pygac.gac_pod import GACPODReader
from pygac.lac_klm import LACKLMReader
from pygac.lac_pod import LACPODReader

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import datetime64_to_pydatetime, get_legacy_chunk_size

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

spacecrafts = {7: "NOAA 15", 3: "NOAA 16", 13: "NOAA 18", 15: "NOAA 19"}

AVHRR3_CHANNEL_NAMES = {"1": 0, "2": 1, "3A": 2, "3B": 3, "4": 4, "5": 5}
AVHRR2_CHANNEL_NAMES = {"1": 0, "2": 1, "3": 2, "4": 3, "5": 4}
AVHRR_CHANNEL_NAMES = {"1": 0, "2": 1, "3": 2, "4": 3}
ANGLES = ("sensor_zenith_angle", "sensor_azimuth_angle", "solar_zenith_angle",
          "solar_azimuth_angle", "sun_sensor_azimuth_difference_angle")


class GACLACFile(BaseFileHandler):
    """Reader for GAC and LAC data."""

    def __init__(self, filename, filename_info, filetype_info,  # noqa: D417
                 start_line=None, end_line=None, strip_invalid_coords=True,
                 interpolate_coords=True, **reader_kwargs):
        """Init the file handler.

        Args:
            start_line: User defined start scanline
            end_line: User defined end scanline
            strip_invalid_coords: Strip scanlines with invalid coordinates in
                the beginning/end of the orbit
            interpolate_coords: Interpolate coordinates from every eighth pixel
                to all pixels.
            reader_kwargs: More keyword arguments to be passed to pygac.Reader.
                See the pygac documentation for available options.

        """
        super(GACLACFile, self).__init__(
            filename, filename_info, filetype_info)

        self.start_line = start_line
        self.end_line = end_line
        self.strip_invalid_coords = strip_invalid_coords
        self.interpolate_coords = interpolate_coords
        self.reader_kwargs = reader_kwargs
        self.creation_site = filename_info.get("creation_site")
        self.reader = None
        self.calib_channels = None
        self.counts = None
        self.angles = None
        self.qual_flags = None
        self.first_valid_lat = None
        self.last_valid_lat = None
        self._start_time = filename_info["start_time"]
        self._end_time = dt.datetime.combine(filename_info["start_time"].date(),
                                             filename_info["end_time"].time())
        if self._end_time < self._start_time:
            self._end_time += dt.timedelta(days=1)
        self.platform_id = filename_info["platform_id"]

        if len(self.platform_id) == 3:
            self.reader_kwargs["header_date"] = dt.date(2000, 1, 1)

        if self._is_avhrr3():
            if filename_info.get("transfer_mode") == "GHRR":
                self.reader_class = GACKLMReader
            else:
                self.reader_class = LACKLMReader
            self.chn_dict = AVHRR3_CHANNEL_NAMES
            self.sensor = "avhrr-3"
        elif self._is_avhrr2():
            if filename_info.get("transfer_mode") == "GHRR":
                self.reader_class = GACPODReader
            else:
                self.reader_class = LACPODReader
            self.chn_dict = AVHRR2_CHANNEL_NAMES
            self.sensor = "avhrr-2"
        else:
            if filename_info.get("transfer_mode") == "GHRR":
                self.reader_class = GACPODReader
            else:
                self.reader_class = LACPODReader
            self.chn_dict = AVHRR_CHANNEL_NAMES
            self.sensor = "avhrr"
        self.filename_info = filename_info

    def _is_avhrr2(self):
        return self.platform_id in ["NC", "NE", "NF", "NG", "NH", "ND", "NJ",
                                    "N07", "N08", "N09", "N10", "N11", "N12", "N14"]

    def _is_avhrr3(self):
        return self.platform_id in ["NK", "NL", "NM", "NN", "NP",
                                    "N15", "N16", "N17", "N18", "N19",
                                    "M1", "M2", "M3",
                                    "MOB", "MOA", "MOC"]

    def read_raw_data(self):
        """Create a pygac reader and read raw data from the file."""
        if self.reader is None:
            self.reader = self.reader_class(
                interpolate_coords=self.interpolate_coords,
                creation_site=self.creation_site,
                **self.reader_kwargs)
            self.reader.read(self.filename)
            if np.all(self.reader.mask):
                raise ValueError("All data is masked out")

    def get_dataset(self, key, info):
        """Get the dataset."""
        self.read_raw_data()
        if key["name"] in ["latitude", "longitude"]:
            # Lats/lons are buffered by the reader
            if key["name"] == "latitude":
                _, data = self.reader.get_lonlat()
            else:
                data, _ = self.reader.get_lonlat()

            # If coordinate interpolation is disabled, only every eighth
            # pixel has a lat/lon coordinate
            xdim = "x" if self.interpolate_coords else "x_every_eighth"
            xcoords = None
        elif key["name"] in ANGLES:
            data = self._get_angle(key)
            xdim = "x" if self.interpolate_coords else "x_every_eighth"
            xcoords = None
        elif key["name"] == "qual_flags":
            data = self.reader.get_qual_flags()
            xdim = "num_flags"
            xcoords = ["Scan line number",
                       "Fatal error flag",
                       "Insufficient data for calibration",
                       "Insufficient data for calibration",
                       "Solar contamination of blackbody in channels 3",
                       "Solar contamination of blackbody in channels 4",
                       "Solar contamination of blackbody in channels 5"]
        elif key["name"].upper() in self.chn_dict:
            # Read and calibrate channel data
            data = self._get_channel(key)
            xdim = "x"
            xcoords = None
        else:
            raise ValueError("Unknown dataset: {}".format(key["name"]))

        # Update start/end time using the actual scanline timestamps
        times = self.reader.get_times()
        self._start_time = datetime64_to_pydatetime(times[0])
        self._end_time = datetime64_to_pydatetime(times[-1])

        # Select user-defined scanlines and/or strip invalid coordinates
        if (self.start_line is not None or self.end_line is not None
                or self.strip_invalid_coords):
            data, times = self.slice(data=data, times=times)

        # Create data array
        chunk_cols = data.shape[1]
        chunk_lines = int((CHUNK_SIZE ** 2) / chunk_cols)
        res = xr.DataArray(da.from_array(data, chunks=(chunk_lines, chunk_cols)),
                           dims=["y", xdim], attrs=info)
        if xcoords:
            res[xdim] = xcoords

        # Update dataset attributes
        self._update_attrs(res)

        # Add scanline acquisition times
        res["acq_time"] = ("y", times)
        res["acq_time"].attrs["long_name"] = "Mean scanline acquisition time"

        return res

    def slice(self, data, times):  # noqa: A003
        """Select user-defined scanlines and/or strip invalid coordinates.

        Furthermore, update scanline timestamps.

        Args:
            data: Data to be sliced
            times: Scanline timestamps
        Returns:
            Sliced data and timestamps

        """
        sliced = self._slice(data)
        times = self._slice(times)
        self._start_time = datetime64_to_pydatetime(times[0])
        self._end_time = datetime64_to_pydatetime(times[-1])
        return sliced, times

    def _slice(self, data):
        """Select user-defined scanlines and/or strip invalid coordinates.

        Returns:
            Sliced data

        """
        start_line = self.start_line if self.start_line is not None else 0
        end_line = self.end_line if self.end_line is not None else 0

        # Strip scanlines with invalid coordinates
        if self.strip_invalid_coords:
            first_valid_lat, last_valid_lat = self._strip_invalid_lat()
        else:
            first_valid_lat = last_valid_lat = None

        # Check and correct user-defined scanlines, if possible
        start_line, end_line = pygac.utils.check_user_scanlines(
            start_line=start_line,
            end_line=end_line,
            first_valid_lat=first_valid_lat,
            last_valid_lat=last_valid_lat,
            along_track=data.shape[0]
        )

        # Slice data
        sliced = pygac.utils.slice_channel(data,
                                           start_line=start_line,
                                           end_line=end_line,
                                           first_valid_lat=first_valid_lat,
                                           last_valid_lat=last_valid_lat)
        if isinstance(sliced, tuple):
            # pygac < 1.4.0
            sliced = sliced[0]

        return sliced

    def _get_channel(self, key):
        """Get channel and buffer results."""
        name = key["name"]
        calibration = key["calibration"]
        if calibration == "counts":
            if self.counts is None:
                counts = self.reader.get_counts()
                self.counts = counts
            channels = self.counts
        elif calibration in ["reflectance", "brightness_temperature"]:
            if self.calib_channels is None:
                self.calib_channels = self.reader.get_calibrated_channels()
            channels = self.calib_channels
        else:
            raise ValueError("Unknown calibration: {}".format(calibration))
        return channels[:, :, self.chn_dict[name.upper()]]

    def _get_qual_flags(self):
        """Get quality flags and buffer results."""
        if self.qual_flags is None:
            self.qual_flags = self.reader.get_qual_flags()
        return self.qual_flags

    def _get_angle(self, key):
        """Get angles and buffer results."""
        if self.angles is None:
            sat_azi, sat_zenith, sun_azi, sun_zenith, rel_azi = self.reader.get_angles()
            self.angles = {"sensor_zenith_angle": sat_zenith,
                           "sensor_azimuth_angle": sat_azi,
                           "solar_zenith_angle": sun_zenith,
                           "solar_azimuth_angle": sun_azi,
                           "sun_sensor_azimuth_difference_angle": rel_azi}
        return self.angles[key["name"]]

    def _strip_invalid_lat(self):
        """Strip scanlines with invalid coordinates in the beginning/end of the orbit.

        Returns:
            First and last scanline with valid latitudes.

        """
        if self.first_valid_lat is None:
            _, lats = self.reader.get_lonlat()
            start, end = pygac.utils.strip_invalid_lat(lats)
            self.first_valid_lat, self.last_valid_lat = start, end
        return self.first_valid_lat, self.last_valid_lat

    def _update_attrs(self, res):
        """Update dataset attributes."""
        for attr in self.reader.meta_data:
            res.attrs[attr] = self.reader.meta_data[attr]
        res.attrs["platform_name"] = self.reader.spacecraft_name
        res.attrs["orbit_number"] = self.filename_info.get("orbit_number", None)
        res.attrs["sensor"] = self.sensor
        try:
            res.attrs["orbital_parameters"] = {"tle": self.reader.get_tle_lines()}
        except (IndexError, RuntimeError):
            pass

    @property
    def start_time(self):
        """Get the start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get the end time."""
        return self._end_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for saving animations using Multiscene."""

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path

import datetime as dt
import os
import shutil
import tempfile
import unittest
from unittest import mock

import pytest

from satpy.tests.multiscene_tests.test_utils import (
    _create_test_area,
    _create_test_dataset,
    _create_test_scenes,
    _fake_get_enhanced_image,
)


class TestMultiSceneSave(unittest.TestCase):
    """Test saving a MultiScene to various formats."""

    def setUp(self):
        """Create temporary directory to save files to."""
        self.base_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Remove the temporary directory created for a test."""
        try:
            shutil.rmtree(self.base_dir, ignore_errors=True)
        except OSError:
            pass

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image", _fake_get_enhanced_image)
    def test_save_mp4_distributed(self):
        """Save a series of fake scenes to an mp4 video."""
        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]["ds3"] = _create_test_dataset("ds3")
        # Add a start and end time
        for ds_id in ["ds1", "ds2", "ds3"]:
            scenes[1][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 2)
            scenes[1][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 2, 12)
            if ds_id == "ds3":
                continue
            scenes[0][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 1)
            scenes[0][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        fn = os.path.join(
            self.base_dir,
            "test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4")
        writer_mock = mock.MagicMock()
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v.compute() for v in x)
        client_mock.gather.side_effect = lambda x: x
        with mock.patch("satpy.multiscene._multiscene.imageio.get_writer") as get_writer:
            get_writer.return_value = writer_mock
            # force order of datasets by specifying them
            mscn.save_animation(fn, client=client_mock, datasets=["ds1", "ds2", "ds3"])

        # 2 saves for the first scene + 1 black frame
        # 3 for the second scene
        assert writer_mock.append_data.call_count == 3 + 3
        filenames = [os.path.basename(args[0][0]) for args in get_writer.call_args_list]
        assert filenames[0] == "test_save_mp4_ds1_20180101_00_20180102_12.mp4"
        assert filenames[1] == "test_save_mp4_ds2_20180101_00_20180102_12.mp4"
        assert filenames[2] == "test_save_mp4_ds3_20180102_00_20180102_12.mp4"

        # Test no distributed client found
        mscn = MultiScene(scenes)
        fn = os.path.join(
            self.base_dir,
            "test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4")
        writer_mock = mock.MagicMock()
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v.compute() for v in x)
        client_mock.gather.side_effect = lambda x: x
        with mock.patch("satpy.multiscene._multiscene.imageio.get_writer") as get_writer, \
                mock.patch("satpy.multiscene._multiscene.get_client", mock.Mock(side_effect=ValueError("No client"))):
            get_writer.return_value = writer_mock
            # force order of datasets by specifying them
            mscn.save_animation(fn, datasets=["ds1", "ds2", "ds3"])

        # 2 saves for the first scene + 1 black frame
        # 3 for the second scene
        assert writer_mock.append_data.call_count == 3 + 3
        filenames = [os.path.basename(args[0][0]) for args in get_writer.call_args_list]
        assert filenames[0] == "test_save_mp4_ds1_20180101_00_20180102_12.mp4"
        assert filenames[1] == "test_save_mp4_ds2_20180101_00_20180102_12.mp4"
        assert filenames[2] == "test_save_mp4_ds3_20180102_00_20180102_12.mp4"

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image", _fake_get_enhanced_image)
    def test_save_mp4_no_distributed(self):
        """Save a series of fake scenes to an mp4 video when distributed isn't available."""
        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]["ds3"] = _create_test_dataset("ds3")
        # Add a start and end time
        for ds_id in ["ds1", "ds2", "ds3"]:
            scenes[1][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 2)
            scenes[1][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 2, 12)
            if ds_id == "ds3":
                continue
            scenes[0][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 1)
            scenes[0][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        fn = os.path.join(
            self.base_dir,
            "test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4")
        writer_mock = mock.MagicMock()
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v.compute() for v in x)
        client_mock.gather.side_effect = lambda x: x
        with mock.patch("satpy.multiscene._multiscene.imageio.get_writer") as get_writer, \
                mock.patch("satpy.multiscene._multiscene.get_client", None):
            get_writer.return_value = writer_mock
            # force order of datasets by specifying them
            mscn.save_animation(fn, datasets=["ds1", "ds2", "ds3"])

        # 2 saves for the first scene + 1 black frame
        # 3 for the second scene
        assert writer_mock.append_data.call_count == 3 + 3
        filenames = [os.path.basename(args[0][0]) for args in get_writer.call_args_list]
        assert filenames[0] == "test_save_mp4_ds1_20180101_00_20180102_12.mp4"
        assert filenames[1] == "test_save_mp4_ds2_20180101_00_20180102_12.mp4"
        assert filenames[2] == "test_save_mp4_ds3_20180102_00_20180102_12.mp4"

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image", _fake_get_enhanced_image)
    def test_save_datasets_simple(self):
        """Save a series of fake scenes to an PNG images."""
        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]["ds3"] = _create_test_dataset("ds3")
        # Add a start and end time
        for ds_id in ["ds1", "ds2", "ds3"]:
            scenes[1][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 2)
            scenes[1][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 2, 12)
            if ds_id == "ds3":
                continue
            scenes[0][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 1)
            scenes[0][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v for v in x)
        client_mock.gather.side_effect = lambda x: x
        with mock.patch("satpy.multiscene._multiscene.Scene.save_datasets") as save_datasets:
            save_datasets.return_value = [True]  # some arbitrary return value
            # force order of datasets by specifying them
            mscn.save_datasets(base_dir=self.base_dir, client=False, datasets=["ds1", "ds2", "ds3"],
                               writer="simple_image")

        # 2 for each scene
        assert save_datasets.call_count == 2

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image", _fake_get_enhanced_image)
    def test_save_datasets_distributed_delayed(self):
        """Test distributed save for writers returning delayed obejcts e.g. simple_image."""
        from dask.delayed import Delayed

        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]["ds3"] = _create_test_dataset("ds3")
        # Add a start and end time
        for ds_id in ["ds1", "ds2", "ds3"]:
            scenes[1][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 2)
            scenes[1][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 2, 12)
            if ds_id == "ds3":
                continue
            scenes[0][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 1)
            scenes[0][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v for v in x)
        client_mock.gather.side_effect = lambda x: x
        future_mock = mock.MagicMock()
        future_mock.__class__ = Delayed
        with mock.patch("satpy.multiscene._multiscene.Scene.save_datasets") as save_datasets:
            save_datasets.return_value = [future_mock]  # some arbitrary return value
            # force order of datasets by specifying them
            mscn.save_datasets(base_dir=self.base_dir, client=client_mock, datasets=["ds1", "ds2", "ds3"],
                               writer="simple_image")

        # 2 for each scene
        assert save_datasets.call_count == 2

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image", _fake_get_enhanced_image)
    def test_save_datasets_distributed_source_target(self):
        """Test distributed save for writers returning sources and targets e.g. geotiff writer."""
        import dask.array as da

        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]["ds3"] = _create_test_dataset("ds3")
        # Add a start and end time
        for ds_id in ["ds1", "ds2", "ds3"]:
            scenes[1][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 2)
            scenes[1][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 2, 12)
            if ds_id == "ds3":
                continue
            scenes[0][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 1)
            scenes[0][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v for v in x)
        client_mock.gather.side_effect = lambda x: x
        source_mock = mock.MagicMock()
        source_mock.__class__ = da.Array
        target_mock = mock.MagicMock()
        with mock.patch("satpy.multiscene._multiscene.Scene.save_datasets") as save_datasets:
            save_datasets.return_value = ([source_mock], [target_mock])  # some arbitrary return value
            # force order of datasets by specifying them
            with pytest.raises(NotImplementedError):
                mscn.save_datasets(base_dir=self.base_dir, client=client_mock, datasets=["ds1", "ds2", "ds3"],
                                   writer="geotiff")

    def test_crop(self):
        """Test the crop method."""
        import numpy as np
        from pyresample.geometry import AreaDefinition
        from xarray import DataArray

        from satpy import MultiScene, Scene
        scene1 = Scene()
        area_extent = (-5570248.477339745, -5561247.267842293, 5567248.074173927,
                       5570248.477339745)
        proj_dict = {"a": 6378169.0, "b": 6356583.8, "h": 35785831.0,
                     "lon_0": 0.0, "proj": "geos", "units": "m"}
        x_size = 3712
        y_size = 3712
        area_def = AreaDefinition(
            "test", "test", "test",
            proj_dict,
            x_size,
            y_size,
            area_extent,
        )
        area_def2 = AreaDefinition(
            "test2", "test2", "test2", proj_dict,
            x_size // 2,
            y_size // 2,
            area_extent,
        )
        scene1["1"] = DataArray(np.zeros((y_size, x_size)))
        scene1["2"] = DataArray(np.zeros((y_size, x_size)), dims=("y", "x"))
        scene1["3"] = DataArray(np.zeros((y_size, x_size)), dims=("y", "x"),
                                attrs={"area": area_def})
        scene1["4"] = DataArray(np.zeros((y_size // 2, x_size // 2)), dims=("y", "x"),
                                attrs={"area": area_def2})
        mscn = MultiScene([scene1])

        # by lon/lat bbox
        new_mscn = mscn.crop(ll_bbox=(-20., -5., 0, 0))
        new_scn1 = list(new_mscn.scenes)[0]
        assert "1" in new_scn1
        assert "2" in new_scn1
        assert "3" in new_scn1
        assert new_scn1["1"].shape == (y_size, x_size)
        assert new_scn1["2"].shape == (y_size, x_size)
        assert new_scn1["3"].shape == (184, 714)
        assert new_scn1["4"].shape == (92, 357)


@mock.patch("satpy.enhancements.enhancer.get_enhanced_image")
def test_save_mp4(smg, tmp_path):
    """Save a series of fake scenes to an mp4 video."""
    from satpy import MultiScene
    area = _create_test_area()
    scenes = _create_test_scenes(area=area)
    smg.side_effect = _fake_get_enhanced_image

    # Add a dataset to only one of the Scenes
    scenes[1]["ds3"] = _create_test_dataset("ds3")
    # Add a start and end time
    for ds_id in ["ds1", "ds2", "ds3"]:
        scenes[1][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 2)
        scenes[1][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 2, 12)
        if ds_id == "ds3":
            continue
        scenes[0][ds_id].attrs["start_time"] = dt.datetime(2018, 1, 1)
        scenes[0][ds_id].attrs["end_time"] = dt.datetime(2018, 1, 1, 12)

    mscn = MultiScene(scenes)
    fn = str(tmp_path /
             "test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4")
    writer_mock = mock.MagicMock()
    with mock.patch("satpy.multiscene._multiscene.imageio.get_writer") as get_writer:
        get_writer.return_value = writer_mock
        # force order of datasets by specifying them
        mscn.save_animation(fn, datasets=["ds1", "ds2", "ds3"], client=False)

    # 2 saves for the first scene + 1 black frame
    # 3 for the second scene
    assert writer_mock.append_data.call_count == 3 + 3
    filenames = [os.path.basename(args[0][0]) for args in get_writer.call_args_list]
    assert filenames[0] == "test_save_mp4_ds1_20180101_00_20180102_12.mp4"
    assert filenames[1] == "test_save_mp4_ds2_20180101_00_20180102_12.mp4"
    assert filenames[2] == "test_save_mp4_ds3_20180102_00_20180102_12.mp4"

    # make sure that not specifying datasets still saves all of them
    fn = str(tmp_path /
             "test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4")
    writer_mock = mock.MagicMock()
    with mock.patch("satpy.multiscene._multiscene.imageio.get_writer") as get_writer:
        get_writer.return_value = writer_mock
        # force order of datasets by specifying them
        mscn.save_animation(fn, client=False)
    # the 'ds3' dataset isn't known to the first scene so it doesn't get saved
    # 2 for first scene, 2 for second scene
    assert writer_mock.append_data.call_count == 2 + 2
    assert "test_save_mp4_ds1_20180101_00_20180102_12.mp4" in filenames
    assert "test_save_mp4_ds2_20180101_00_20180102_12.mp4" in filenames
    assert "test_save_mp4_ds3_20180102_00_20180102_12.mp4" in filenames

    # test decorating and enhancing

    fn = str(tmp_path /
             "test-{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}-rich.mp4")
    writer_mock = mock.MagicMock()
    with mock.patch("satpy.multiscene._multiscene.imageio.get_writer") as get_writer:
        get_writer.return_value = writer_mock
        mscn.save_animation(
            fn, client=False,
            enh_args={"decorate": {
                "decorate": [{
                    "text": {
                        "txt":
                            "Test {start_time:%Y-%m-%d %H:%M} - "
                            "{end_time:%Y-%m-%d %H:%M}"}}]}})
    assert writer_mock.append_data.call_count == 2 + 2
    assert ("2018-01-02" in smg.call_args_list[-1][1]
            ["decorate"]["decorate"][0]["text"]["txt"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Composite classes for the AGRI instrument."""

import logging

from .core import GenericCompositor

LOG = logging.getLogger(__name__)


class SimulatedRed(GenericCompositor):
    """A single-band dataset resembling a Red (0.64 µm) band.

    This compositor creates a single band product by combining two
    other bands by preset amounts. The general formula with
    dependencies (d) and fractions (f) is::

        result = (f1 * d1 - f2 * d2) / f3

    See the `fractions` keyword argument for more information.
    The default setup is to use:

     - f1 = 1.0
     - f2 = 0.13
     - f3 = 0.87

    """

    def __init__(self, name, fractions=(1.0, 0.13, 0.87), **kwargs):  # noqa: D417
        """Initialize fractions for input channels.

        Args:
            name (str): Name of this composite
            fractions (Iterable): Fractions of each input band to include in the result.

        """
        self.fractions = fractions
        super(SimulatedRed, self).__init__(name, **kwargs)

    def __call__(self, projectables, optional_datasets=None, **attrs):
        """Generate the single band composite."""
        c1, c2 = self.match_data_arrays(projectables)
        res = (c1 * self.fractions[0] - c2 * self.fractions[1]) / self.fractions[2]
        res.attrs = c1.attrs.copy()
        return super(SimulatedRed, self).__call__((res,), **attrs)



# Copyright (c) 2019-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Testing of utils."""
from __future__ import annotations

import datetime
import logging
import typing
import warnings
from math import sqrt
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.utils import (
    angle2xyz,
    datetime64_to_pydatetime,
    get_legacy_chunk_size,
    get_satpos,
    import_error_helper,
    lonlat2xyz,
    proj_units_to_meters,
    xyz2angle,
    xyz2lonlat,
)

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - caplog


class TestGeoUtils:
    """Testing geo-related utility functions."""

    @pytest.mark.parametrize(
        ("lonlat", "xyz"),
        [
            ((0, 0), (1, 0, 0)),
            ((90, 0), (0, 1, 0)),
            ((0, 90), (0, 0, 1)),
            ((180, 0), (-1, 0, 0)),
            ((-90, 0), (0, -1, 0)),
            ((0, -90), (0, 0, -1)),
            ((0, 45), (sqrt(2) / 2, 0, sqrt(2) / 2)),
            ((0, 60), (sqrt(1) / 2, 0, sqrt(3) / 2)),
        ],
    )
    def test_lonlat2xyz(self, lonlat, xyz):
        """Test the lonlat2xyz function."""
        x__, y__, z__ = lonlat2xyz(*lonlat)
        assert x__ == pytest.approx(xyz[0])
        assert y__ == pytest.approx(xyz[1])
        assert z__ == pytest.approx(xyz[2])

    @pytest.mark.parametrize(
        ("azizen", "xyz"),
        [
            ((0, 0), (0, 0, 1)),
            ((90, 0), (0, 0, 1)),
            ((0, 90), (0, 1, 0)),
            ((180, 0), (0, 0, 1)),
            ((-90, 0), (0, 0, 1)),
            ((0, -90), (0, -1, 0)),
            ((90, 90), (1, 0, 0)),
            ((-90, 90), (-1, 0, 0)),
            ((180, 90), (0, -1, 0)),
            ((0, 45), (0, sqrt(2) / 2, sqrt(2) / 2)),
            ((0, 60), (0, sqrt(3) / 2, sqrt(1) / 2)),
        ],
    )
    def test_angle2xyz(self, azizen, xyz):
        """Test the angle2xyz function."""
        x__, y__, z__ = angle2xyz(*azizen)
        assert x__ == pytest.approx(xyz[0])
        assert y__ == pytest.approx(xyz[1])
        assert z__ == pytest.approx(xyz[2])

    @pytest.mark.parametrize(
        ("xyz", "asin", "lonlat"),
        [
            ((1, 0, 0), False, (0, 0)),
            ((0, 1, 0), False, (90, 0)),
            ((0, 0, 1), True, (0, 90)),
            ((0, 0, 1), False, (0, 90)),
            ((sqrt(2) / 2, sqrt(2) / 2, 0), False, (45, 0)),
        ],
    )
    def test_xyz2lonlat(self, xyz, asin, lonlat):
        """Test xyz2lonlat."""
        lon, lat = xyz2lonlat(*xyz, asin=asin)
        assert lon == pytest.approx(lonlat[0])
        assert lat == pytest.approx(lonlat[1])

    @pytest.mark.parametrize(
        ("xyz", "acos", "azizen"),
        [
            ((1, 0, 0), False, (90, 90)),
            ((0, 1, 0), False, (0, 90)),
            ((0, 0, 1), False, (0, 0)),
            ((0, 0, 1), True, (0, 0)),
            ((sqrt(2) / 2, sqrt(2) / 2, 0), False, (45, 90)),
            ((-1, 0, 0), False, (-90, 90)),
            ((0, -1, 0), False, (180, 90)),
        ],
    )
    def test_xyz2angle(self, xyz, acos, azizen):
        """Test xyz2angle."""
        azi, zen = xyz2angle(*xyz, acos=acos)
        assert azi == pytest.approx(azi)
        assert zen == pytest.approx(zen)

    @pytest.mark.parametrize(
        ("prj", "exp_prj"),
        [
            ("+asd=123123123123", "+asd=123123123123"),
            ("+a=6378.137", "+a=6378137.000"),
            ("+a=6378.137 +units=km", "+a=6378137.000"),
            ("+a=6378.137 +b=6378.137", "+a=6378137.000 +b=6378137.000"),
            ("+a=6378.137 +b=6378.137 +h=35785.863", "+a=6378137.000 +b=6378137.000 +h=35785863.000"),
        ],
    )
    def test_proj_units_to_meters(self, prj, exp_prj):
        """Test proj units to meters conversion."""
        assert proj_units_to_meters(prj) == exp_prj


class TestGetSatPos:
    """Tests for 'get_satpos'."""

    @pytest.mark.parametrize(
        ("included_prefixes", "preference", "expected_result"),
        [
            (("nadir_", "satellite_actual_", "satellite_nominal_", "projection_"), None, (1, 2, 3)),
            (("satellite_actual_", "satellite_nominal_", "projection_"), None, (1.1, 2.1, 3)),
            (("satellite_nominal_", "projection_"), None, (1.2, 2.2, 3.1)),
            (("projection_",), None, (1.3, 2.3, 3.2)),
            (("nadir_", "satellite_actual_", "satellite_nominal_", "projection_"), "nadir", (1, 2, 3)),
            (("nadir_", "satellite_actual_", "satellite_nominal_", "projection_"), "actual", (1.1, 2.1, 3)),
            (("nadir_", "satellite_actual_", "satellite_nominal_", "projection_"), "nominal", (1.2, 2.2, 3.1)),
            (("nadir_", "satellite_actual_", "satellite_nominal_", "projection_"), "projection", (1.3, 2.3, 3.2)),
            (("satellite_nominal_", "projection_"), "actual", (1.2, 2.2, 3.1)),
            (("projection_",), "projection", (1.3, 2.3, 3.2)),
        ]
    )
    def test_get_satpos(self, included_prefixes, preference, expected_result):
        """Test getting the satellite position."""
        all_orb_params = {
            "nadir_longitude": 1,
            "satellite_actual_longitude": 1.1,
            "satellite_nominal_longitude": 1.2,
            "projection_longitude": 1.3,
            "nadir_latitude": 2,
            "satellite_actual_latitude": 2.1,
            "satellite_nominal_latitude": 2.2,
            "projection_latitude": 2.3,
            "satellite_actual_altitude": 3,
            "satellite_nominal_altitude": 3.1,
            "projection_altitude": 3.2
        }
        orb_params = {key: value for key, value in all_orb_params.items() if
                      any(in_prefix in key for in_prefix in included_prefixes)}
        data_arr = xr.DataArray((), attrs={"orbital_parameters": orb_params})

        with warnings.catch_warnings(record=True) as caught_warnings:
            lon, lat, alt = get_satpos(data_arr, preference=preference)
        has_satpos_warnings = any("using projection" in str(msg.message) for msg in caught_warnings)
        expect_warning = included_prefixes == ("projection_",) and preference != "projection"
        if expect_warning:
            assert has_satpos_warnings
        else:
            assert not has_satpos_warnings
        assert (lon, lat, alt) == expected_result

    @pytest.mark.parametrize(
        "attrs",
        [
                {},
                {"orbital_parameters": {"projection_longitude": 1}},
                {"satellite_altitude": 1}
        ]
    )
    def test_get_satpos_fails_with_informative_error(self, attrs):
        """Test that get_satpos raises an informative error message."""
        data_arr = xr.DataArray((), attrs=attrs)
        with pytest.raises(KeyError, match="Unable to determine satellite position.*"):
            get_satpos(data_arr)

    def test_get_satpos_from_satname(self, caplog):
        """Test getting satellite position from satellite name only."""
        import pyorbital.tlefile

        data_arr = xr.DataArray(
            (),
            attrs={
                "platform_name": "Meteosat-42",
                "sensor": "irives",
                "start_time": datetime.datetime(2031, 11, 20, 19, 18, 17)
            })
        with mock.patch("pyorbital.tlefile.read") as plr:
            plr.return_value = pyorbital.tlefile.Tle(
                "Meteosat-42",
                line1="1 40732U 15034A   22011.84285506  .00000004  00000+0  00000+0 0  9995",
                line2="2 40732   0.2533 325.0106 0000976 118.8734 330.4058  1.00272123 23817")
            with caplog.at_level(logging.WARNING):
                (lon, lat, alt) = get_satpos(data_arr, use_tle=True)
            assert "Orbital parameters missing from metadata" in caplog.text
            np.testing.assert_allclose(
                (lon, lat, alt),
                (119.39533705010592, -1.1491628298731498, 35803199.86408156),
                rtol=1e-4,
            )


def test_make_fake_scene():
    """Test the make_fake_scene utility.

    Although the make_fake_scene utility is for internal testing
    purposes, it has grown sufficiently complex that it needs its own
    testing.
    """
    from satpy.tests.utils import make_fake_scene

    assert make_fake_scene({}).keys() == []
    sc = make_fake_scene({
        "six": np.arange(25).reshape(5, 5)
    })
    assert len(sc.keys()) == 1
    assert sc.keys().pop()["name"] == "six"
    assert sc["six"].attrs["area"].shape == (5, 5)
    sc = make_fake_scene({
        "seven": np.arange(3 * 7).reshape(3, 7),
        "eight": np.arange(3 * 8).reshape(3, 8)
    },
        daskify=True,
        area=False,
        common_attrs={"repetency": "fourteen hundred per centimetre"})
    assert "area" not in sc["seven"].attrs.keys()
    assert (sc["seven"].attrs["repetency"] == sc["eight"].attrs["repetency"] ==
            "fourteen hundred per centimetre")
    assert isinstance(sc["seven"].data, da.Array)
    sc = make_fake_scene({
        "nine": xr.DataArray(
            np.arange(2 * 9).reshape(2, 9),
            dims=("y", "x"),
            attrs={"please": "preserve", "answer": 42})
    },
        common_attrs={"bad words": "semprini bahnhof veerooster winterbanden"})
    assert sc["nine"].attrs.keys() >= {"please", "answer", "bad words", "area"}


class TestCheckSatpy:
    """Test the 'check_satpy' function."""

    def test_basic_check_satpy(self):
        """Test 'check_satpy' basic functionality."""
        from satpy.utils import check_satpy
        check_satpy()

    def test_specific_check_satpy(self, capsys):
        """Test 'check_satpy' with specific features provided."""
        from satpy.utils import check_satpy
        check_satpy(readers=["viirs_sdr"], packages=("cartopy", "__fake"))
        out, _ = capsys.readouterr()
        checked_fake = "__fake: not installed" in out
        checked_viirs_sdr = "Readers\n=======\nviirs_sdr" in out
        assert checked_fake, "Did not find __fake package mentioned in checks"
        assert checked_viirs_sdr, "Did not find viirs_sdr in readers mentioned in checks"


class TestShowVersions:
    """Test the 'show_versions' function."""

    def test_basic_show_versions(self):
        """Test 'check_satpy' basic functionality."""
        from satpy.utils import show_versions
        show_versions()

    def test_show_specific_version(self, capsys):
        """Test 'show_version' works with installed package."""
        from satpy.utils import show_versions
        show_versions(packages=["pytest"])
        out, _ = capsys.readouterr()

        pytest_mentioned = "pytest:" in out
        pytest_installed = "pytest: not installed" not in out
        check_pytest = pytest_mentioned and pytest_installed
        assert check_pytest, "pytest with package version not in print output"

    def test_show_missing_specific_version(self, capsys):
        """Test 'show_version' works with missing package."""
        from satpy.utils import show_versions
        show_versions(packages=["__fake"])
        out, _ = capsys.readouterr()

        check_fake = "__fake: not installed" in out
        assert check_fake, "Did not find '__fake: not installed' in print output"


def test_debug_on(caplog):
    """Test that debug_on is working as expected."""
    from satpy.utils import debug, debug_off, debug_on

    def depwarn():
        logger = logging.getLogger("satpy.silly")
        logger.debug("But now it's just got SILLY.")
        warnings.warn("Stop that! It's SILLY.", DeprecationWarning, stacklevel=2)
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    debug_on(False)
    filts_before = warnings.filters.copy()
    # test that logging on, but deprecation warnings still off
    with caplog.at_level(logging.DEBUG):
        depwarn()
    assert warnings.filters == filts_before
    assert "But now it's just got SILLY." in caplog.text
    debug_on(True)
    # test that logging on and deprecation warnings on
    with pytest.warns(DeprecationWarning, match="Stop that! It's SILLY."):
        depwarn()
    assert warnings.filters != filts_before
    debug_off()  # other tests assume debugging is off
    # test that filters were reset
    assert warnings.filters == filts_before
    with debug():
        assert warnings.filters != filts_before
    assert warnings.filters == filts_before


def test_logging_on_and_off(caplog):
    """Test that switching logging on and off works."""
    from satpy.utils import logging_off, logging_on
    logger = logging.getLogger("satpy.silly")
    logging_on()
    with caplog.at_level(logging.WARNING):
        logger.debug("I'd like to leave the army please, sir.")
        logger.warning("Stop that!  It's SILLY.")
    assert "Stop that!  It's SILLY" in caplog.text
    assert "I'd like to leave the army please, sir." not in caplog.text
    logging_off()
    with caplog.at_level(logging.DEBUG):
        logger.warning("You've got a nice army base here, Colonel.")
    assert "You've got a nice army base here, Colonel." not in caplog.text


@pytest.mark.parametrize(
    ("shapes", "chunks", "dims", "exp_unified"),
    [
        (
                ((3, 5, 5), (5, 5)),
                (-1, -1),
                (("bands", "y", "x"), ("y", "x")),
                True,
        ),
        (
                ((3, 5, 5), (5, 5)),
                (-1, 2),
                (("bands", "y", "x"), ("y", "x")),
                True,
        ),
        (
                ((4, 5, 5), (3, 5, 5)),
                (-1, -1),
                (("bands", "y", "x"), ("bands", "y", "x")),
                False,
        ),
    ],
)
def test_unify_chunks(shapes, chunks, dims, exp_unified):
    """Test unify_chunks utility function."""
    from satpy.utils import unify_chunks
    inputs = list(_data_arrays_from_params(shapes, chunks, dims))
    results = unify_chunks(*inputs)
    if exp_unified:
        _verify_unified(results)
    else:
        _verify_unchanged_chunks(results, inputs)


def _data_arrays_from_params(shapes: list[tuple[int, ...]],
                             chunks: list[tuple[int, ...]],
                             dims: list[tuple[int, ...]]
                             ) -> typing.Generator[xr.DataArray, None, None]:
    for shape, chunk, dim in zip(shapes, chunks, dims):
        yield xr.DataArray(da.ones(shape, chunks=chunk), dims=dim)


def _verify_unified(data_arrays: list[xr.DataArray]) -> None:
    dim_chunks: dict[str, int] = {}
    for data_arr in data_arrays:
        for dim, chunk_size in zip(data_arr.dims, data_arr.chunks):
            exp_chunks = dim_chunks.setdefault(dim, chunk_size)
            assert exp_chunks == chunk_size


def _verify_unchanged_chunks(data_arrays: list[xr.DataArray],
                             orig_arrays: list[xr.DataArray]) -> None:
    for data_arr, orig_arr in zip(data_arrays, orig_arrays):
        assert data_arr.chunks == orig_arr.chunks


def test_chunk_size_limit():
    """Check the chunk size limit computations."""
    from unittest.mock import patch

    from satpy.utils import get_chunk_size_limit
    with patch("satpy.utils._get_pytroll_chunk_size") as ptc:
        ptc.return_value = 10
        assert get_chunk_size_limit(np.int32) == 400
        assert get_chunk_size_limit() == 800


def test_chunk_size_limit_from_dask_config():
    """Check the chunk size limit computations."""
    import dask.config

    from satpy.utils import get_chunk_size_limit
    with dask.config.set({"array.chunk-size": "1KiB"}):
        assert get_chunk_size_limit(np.uint8) == 1024


def test_get_legacy_chunk_size():
    """Test getting the legacy chunk size."""
    import dask.config
    assert get_legacy_chunk_size() == 4096
    with dask.config.set({"array.chunk-size": "32MiB"}):
        assert get_legacy_chunk_size() == 2048


@pytest.mark.parametrize(
    ("chunks", "shape", "previous_chunks", "lr_mult", "chunk_dtype", "exp_result"),
    [
        # 1km swath
        (("auto", -1), (1000, 3200), (40, 40), (4, 4), np.float32, (160, -1)),
        # 5km swath
        (("auto", -1), (1000 // 5, 3200 // 5), (40, 40), (20, 20), np.float32, (160 / 5, -1)),
        # 250m swath
        (("auto", -1), (1000 * 4, 3200 * 4), (40, 40), (1, 1), np.float32, (160 * 4, -1)),
        # 1km area (ABI chunk 226):
        (("auto", "auto"), (21696 // 2, 21696 // 2), (226*4, 226*4), (2, 2), np.float32, (1356, 1356)),
        # 1km area (64-bit)
        (("auto", "auto"), (21696 // 2, 21696 // 2), (226*4, 226*4), (2, 2), np.float64, (904, 904)),
        # 3km area
        (("auto", "auto"), (21696 // 3, 21696 // 3), (226*4, 226*4), (6, 6), np.float32, (452, 452)),
        # 500m area
        (("auto", "auto"), (21696, 21696), (226*4, 226*4), (1, 1), np.float32, (1356 * 2, 1356 * 2)),
        # 500m area (64-bit)
        (("auto", "auto"), (21696, 21696), (226*4, 226*4), (1, 1), np.float64, (904 * 2, 904 * 2)),
        # 250m swath with bands:
        ((1, "auto", -1), (7, 1000 * 4, 3200 * 4), (1, 40, 40), (1, 1, 1), np.float32, (1, 160 * 4, -1)),
        # lots of dimensions:
        ((1, 1, "auto", -1), (1, 7, 1000, 3200), (1, 1, 40, 40), (1, 1, 1, 1), np.float32, (1, 1, 1000, -1)),
    ],
)
def test_resolution_chunking(chunks, shape, previous_chunks, lr_mult, chunk_dtype, exp_result):
    """Test normalize_low_res_chunks helper function."""
    import dask.config

    from satpy.utils import normalize_low_res_chunks

    with dask.config.set({"array.chunk-size": "32MiB"}):
        chunk_results = normalize_low_res_chunks(
            chunks,
            shape,
            previous_chunks,
            lr_mult,
            chunk_dtype,
        )
    assert chunk_results == exp_result
    for chunk_size in chunk_results:
        assert isinstance(chunk_size[0], int) if isinstance(chunk_size, tuple) else isinstance(chunk_size, int)

    # make sure the chunks are understandable by dask
    da.zeros(shape, dtype=chunk_dtype, chunks=chunk_results)


def test_convert_remote_files_to_fsspec_local_files():
    """Test convertion of remote files to fsspec objects.

    Case without scheme/protocol, which should default to plain filenames.
    """
    from satpy.utils import convert_remote_files_to_fsspec

    filenames = ["/tmp/file1.nc", "file:///tmp/file2.nc"]
    res = convert_remote_files_to_fsspec(filenames)
    assert res == filenames


def test_convert_remote_files_to_fsspec_local_pathlib_files():
    """Test convertion of remote files to fsspec objects.

    Case using pathlib objects as filenames.
    """
    import pathlib

    from satpy.utils import convert_remote_files_to_fsspec

    filenames = [pathlib.Path("/tmp/file1.nc"), pathlib.Path("c:\tmp\file2.nc")]
    res = convert_remote_files_to_fsspec(filenames)
    assert res == filenames


def test_convert_remote_files_to_fsspec_mixed_sources():
    """Test convertion of remote files to fsspec objects.

    Case with mixed local and remote files.
    """
    from satpy.readers.core.remote import FSFile
    from satpy.utils import convert_remote_files_to_fsspec

    filenames = ["/tmp/file1.nc", "s3://data-bucket/file2.nc", "file:///tmp/file3.nc"]
    res = convert_remote_files_to_fsspec(filenames)
    # Two local files, one remote
    assert filenames[0] in res
    assert filenames[2] in res
    assert sum([isinstance(f, FSFile) for f in res]) == 1


def test_convert_remote_files_to_fsspec_filename_dict():
    """Test convertion of remote files to fsspec objects.

    Case where filenames is a dictionary mapping readers and filenames.
    """
    from satpy.readers.core.remote import FSFile
    from satpy.utils import convert_remote_files_to_fsspec

    filenames = {
        "reader1": ["/tmp/file1.nc", "/tmp/file2.nc"],
        "reader2": ["s3://tmp/file3.nc", "file:///tmp/file4.nc", "/tmp/file5.nc"]
    }
    res = convert_remote_files_to_fsspec(filenames)

    assert res["reader1"] == filenames["reader1"]
    assert filenames["reader2"][1] in res["reader2"]
    assert filenames["reader2"][2] in res["reader2"]
    assert sum([isinstance(f, FSFile) for f in res["reader2"]]) == 1


def test_convert_remote_files_to_fsspec_fsfile():
    """Test convertion of remote files to fsspec objects.

    Case where the some of the files are already FSFile objects.
    """
    from satpy.readers.core.remote import FSFile
    from satpy.utils import convert_remote_files_to_fsspec

    filenames = ["/tmp/file1.nc", "s3://data-bucket/file2.nc", FSFile("ssh:///tmp/file3.nc")]
    res = convert_remote_files_to_fsspec(filenames)

    assert sum([isinstance(f, FSFile) for f in res]) == 2


def test_convert_remote_files_to_fsspec_windows_paths():
    """Test convertion of remote files to fsspec objects.

    Case where windows paths are used.
    """
    from satpy.utils import convert_remote_files_to_fsspec

    filenames = [r"C:\wintendo\file1.nc", "e:\\wintendo\\file2.nc", r"wintendo\file3.nc"]
    res = convert_remote_files_to_fsspec(filenames)

    assert res == filenames


@mock.patch("fsspec.open_files")
def test_convert_remote_files_to_fsspec_storage_options(open_files):
    """Test convertion of remote files to fsspec objects.

    Case with storage options given.
    """
    from satpy.utils import convert_remote_files_to_fsspec

    filenames = ["s3://tmp/file1.nc"]
    storage_options = {"anon": True}

    _ = convert_remote_files_to_fsspec(filenames, storage_options=storage_options)

    open_files.assert_called_once_with(filenames, **storage_options)


def test_import_error_helper():
    """Test the import error helper."""
    module = "some_crazy_name_for_unknow_dependency_module"
    with pytest.raises(ImportError) as err:  # noqa: PT012
        with import_error_helper(module):
            import unknow_dependency_module  # noqa
    assert module in str(err)


def test_find_in_ancillary():
    """Test finding a dataset in ancillary variables."""
    from satpy.utils import find_in_ancillary
    index_finger = xr.DataArray(
        data=np.arange(25).reshape(5, 5),
        dims=("y", "x"),
        attrs={"name": "index-finger"})
    ring_finger = xr.DataArray(
        data=np.arange(25).reshape(5, 5),
        dims=("y", "x"),
        attrs={"name": "ring-finger"})

    hand = xr.DataArray(
        data=np.arange(25).reshape(5, 5),
        dims=("y", "x"),
        attrs={
            "name": "hand",
            "ancillary_variables": [index_finger, index_finger, ring_finger]
        })

    assert find_in_ancillary(hand, "ring-finger") is ring_finger
    with pytest.raises(
            ValueError,
            match=("Expected exactly one dataset named index-finger in "
                   "ancillary variables for dataset 'hand', found 2")):
        find_in_ancillary(hand, "index-finger")
    with pytest.raises(
            ValueError,
            match=("Could not find dataset named thumb in "
                   "ancillary variables for dataset 'hand'")):
        find_in_ancillary(hand, "thumb")


@pytest.mark.parametrize(
    ("dt64", "expected"),
    [
        (
                np.datetime64("2000-01-02T03:04:05.000000006"),
                datetime.datetime(2000, 1, 2, 3, 4, 5, 0)
        ),
        (
                np.datetime64("2000-01-02T03:04:05.000006"),
                datetime.datetime(2000, 1, 2, 3, 4, 5, 6)
        )
    ]
)
def test_datetime64_to_pydatetime(dt64, expected):
    """Test conversion from datetime64 to Python datetime."""
    assert datetime64_to_pydatetime(dt64) == expected


def test_flatten_dict():
    """Test dictionary flattening."""
    from satpy.utils import flatten_dict
    d = {"a": 1, "b": {"c": 1, "d": {"e": 1, "f": {"g": [1, 2]}}}}
    expected = {"a": 1,
                "b_c": 1,
                "b_d_e": 1,
                "b_d_f_g": [1, 2]}
    assert flatten_dict(d) == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Common functionality for SEVIRI L1.5 data readers.

Calibration
-----------

This section describes how to control the calibration of SEVIRI L1.5 data.


Calibration to radiance
^^^^^^^^^^^^^^^^^^^^^^^

The SEVIRI L1.5 data readers allow for choosing between two file-internal
calibration coefficients to convert counts to radiances:

    - Nominal for all channels (default)
    - GSICS where available (IR currently) and nominal for the remaining
      channels (VIS & HRV currently)

In order to change the default behaviour, use the ``reader_kwargs`` keyword
argument upon Scene creation::

    import satpy
    scene = satpy.Scene(filenames=filenames,
                        reader='seviri_l1b_...',
                        reader_kwargs={'calib_mode': 'GSICS'})
    scene.load(['VIS006', 'IR_108'])

In addition, two other calibration methods are available:

1. It is possible to specify external calibration coefficients for the
   conversion from counts to radiances. External coefficients take
   precedence over internal coefficients and over the Meirink
   coefficients, but you can also mix internal and external coefficients:
   If external calibration coefficients are specified for only a subset
   of channels, the remaining channels will be calibrated using the
   chosen file-internal coefficients (nominal or GSICS).  Calibration
   coefficients must be specified in [mW m-2 sr-1 (cm-1)-1].

2. The calibration mode ``meirink-2023`` uses coefficients based on an
   intercalibration with Aqua-MODIS for the visible channels, as found in
   `Inter-calibration of polar imager solar channels using SEVIRI`_
   (2013) by J. F. Meirink, R. A. Roebeling, and P. Stammes.


In the following example we use external calibration coefficients for the
``VIS006`` & ``IR_108`` channels, and nominal coefficients for the
remaining channels::

    coefs = {'VIS006': {'gain': 0.0236, 'offset': -1.20},
             'IR_108': {'gain': 0.2156, 'offset': -10.4}}
    scene = satpy.Scene(filenames,
                        reader='seviri_l1b_...',
                        reader_kwargs={'ext_calib_coefs': coefs})
    scene.load(['VIS006', 'VIS008', 'IR_108', 'IR_120'])

In the next example we use external calibration coefficients for the
``VIS006`` & ``IR_108`` channels, GSICS coefficients where available
(other IR channels) and nominal coefficients for the rest::

    coefs = {'VIS006': {'gain': 0.0236, 'offset': -1.20},
             'IR_108': {'gain': 0.2156, 'offset': -10.4}}
    scene = satpy.Scene(filenames,
                        reader='seviri_l1b_...',
                        reader_kwargs={'calib_mode': 'GSICS',
                                       'ext_calib_coefs': coefs})
    scene.load(['VIS006', 'VIS008', 'IR_108', 'IR_120'])

In the next example we use the mode ``meirink-2023`` calibration
coefficients for all visible channels and nominal coefficients for the
rest::

    scene = satpy.Scene(filenames,
                        reader='seviri_l1b_...',
                        reader_kwargs={'calib_mode': 'meirink-2023'})
    scene.load(['VIS006', 'VIS008', 'IR_016'])


Calibration to reflectance
^^^^^^^^^^^^^^^^^^^^^^^^^^

When loading solar channels, the SEVIRI L1.5 data readers apply a correction for
the Sun-Earth distance variation throughout the year - as recommended by
the EUMETSAT document
`Conversion from radiances to reflectances for SEVIRI warm channels`_.
In the unlikely situation that this correction is not required, it can be
removed on a per-channel basis using
:func:`satpy.readers.core.utils.remove_earthsun_distance_correction`.


Masking of bad quality scan lines
---------------------------------

By default bad quality scan lines are masked and replaced with ``np.nan`` for radiance, reflectance and
brightness temperature calibrations based on the quality flags provided by the data (for details on quality
flags see `MSG Level 1.5 Image Data Format Description`_ page 109). To disable masking
``reader_kwargs={'mask_bad_quality_scan_lines': False}`` can be passed to the Scene.


Metadata
--------

The SEVIRI L1.5 readers provide the following metadata:

* The ``orbital_parameters`` attribute provides the nominal and actual satellite
  position, as well as the projection centre. See the `Metadata` section in
  the :doc:`../reading` chapter for more information.

* The ``acq_time`` coordinate provides the mean acquisition time for each
  scanline. Use a ``MultiIndex`` to enable selection by acquisition time:

  .. code-block:: python

      import pandas as pd
      mi = pd.MultiIndex.from_arrays([scn['IR_108']['y'].data, scn['IR_108']['acq_time'].data],
                                     names=('y_coord', 'time'))
      scn['IR_108']['y'] = mi
      scn['IR_108'].sel(time=np.datetime64('2019-03-01T12:06:13.052000000'))

* HRIT and Native readers can add raw metadata from the file header, such
  as calibration coefficients, to dataset attributes. Use the reader keyword
  argument ``include_raw_metadata``. Here's an example for extracting
  calibration coefficients from Native files.

  .. code-block:: python

       scene = satpy.Scene(filenames,
                           reader='seviri_l1b_native',
                           reader_kwargs={'include_raw_metadata': True})
       scene.load(["IR_108"])
       mda = scene["IR_108"].attrs["raw_metadata"]
       coefs = mda["15_DATA_HEADER"]["RadiometricProcessing"]["Level15ImageCalibration"]

  Note that this comes with a performance penalty of up to 10% if raw metadata from
  multiple segments or scans need to be combined. By default, arrays with more
  than 100 elements are excluded to limit the performance penalty. This
  threshold can be adjusted using the ``mda_max_array_size`` reader keyword
  argument:

  .. code-block:: python

       scene = satpy.Scene(filenames,
                           reader='seviri_l1b_native',
                           reader_kwargs={'include_raw_metadata': True,
                                         'mda_max_array_size': 1000})

References:
    - `MSG Level 1.5 Image Data Format Description`_
    - `Radiometric Calibration of MSG SEVIRI Level 1.5 Image Data in Equivalent Spectral Blackbody Radiance`_

.. _Conversion from radiances to reflectances for SEVIRI warm channels:
    https://user.eumetsat.int/s3/eup-strapi-media/pdf_msg_seviri_rad2refl_d6cd0c663f.pdf

.. _MSG Level 1.5 Image Data Format Description:
    https://user.eumetsat.int/s3/eup-strapi-media/pdf_ten_05105_msg_img_data_e7c8b315e6.pdf

.. _Radiometric Calibration of MSG SEVIRI Level 1.5 Image Data in Equivalent Spectral Blackbody Radiance:
    https://user.eumetsat.int/s3/eup-strapi-media/pdf_ten_msg_seviri_rad_calib_b4e474a6de.pdf

.. _Inter-calibration of polar imager solar channels using SEVIRI:
   http://dx.doi.org/10.5194/amt-6-2495-2013

"""
from __future__ import annotations

import datetime as dt
import warnings
from collections import namedtuple
from collections.abc import Iterable, Sequence
from enum import IntEnum

import dask.array as da
import numpy as np
import pyproj
import xarray as xr
from numpy.polynomial.chebyshev import Chebyshev

import satpy.readers.core.utils as utils
from satpy.readers.core.eum import issue_revision, time_cds_short
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()
PLATFORM_DICT = {
    "MET08": "Meteosat-8",
    "MET09": "Meteosat-9",
    "MET10": "Meteosat-10",
    "MET11": "Meteosat-11",
    "MSG1": "Meteosat-8",
    "MSG2": "Meteosat-9",
    "MSG3": "Meteosat-10",
    "MSG4": "Meteosat-11",
}

REPEAT_CYCLE_DURATION = 15

REPEAT_CYCLE_DURATION_RSS = 5

C1 = 1.19104273e-5
C2 = 1.43877523

VISIR_NUM_COLUMNS = 3712
VISIR_NUM_LINES = 3712
HRV_NUM_COLUMNS = 11136
HRV_NUM_LINES = 11136

CHANNEL_NAMES = {1: "VIS006",
                 2: "VIS008",
                 3: "IR_016",
                 4: "IR_039",
                 5: "WV_062",
                 6: "WV_073",
                 7: "IR_087",
                 8: "IR_097",
                 9: "IR_108",
                 10: "IR_120",
                 11: "IR_134",
                 12: "HRV"}

VIS_CHANNELS = ["HRV", "VIS006", "VIS008", "IR_016"]

# Polynomial coefficients for spectral-effective BT fits
BTFIT = dict()
# [A, B, C]
BTFIT["IR_039"] = [0.0, 1.011751900, -3.550400]
BTFIT["WV_062"] = [0.00001805700, 1.000255533, -1.790930]
BTFIT["WV_073"] = [0.00000231818, 1.000668281, -0.456166]
BTFIT["IR_087"] = [-0.00002332000, 1.011803400, -1.507390]
BTFIT["IR_097"] = [-0.00002055330, 1.009370670, -1.030600]
BTFIT["IR_108"] = [-0.00007392770, 1.032889800, -3.296740]
BTFIT["IR_120"] = [-0.00007009840, 1.031314600, -3.181090]
BTFIT["IR_134"] = [-0.00007293450, 1.030424800, -2.645950]

SATNUM = {321: "8",
          322: "9",
          323: "10",
          324: "11"}

CALIB = dict()

# Meteosat 8
CALIB[321] = {"HRV": {"F": 78.7599},
              "VIS006": {"F": 65.2296},
              "VIS008": {"F": 73.0127},
              "IR_016": {"F": 62.3715},
              "IR_039": {"VC": 2567.33,
                         "ALPHA": 0.9956,
                         "BETA": 3.41},
              "WV_062": {"VC": 1598.103,
                         "ALPHA": 0.9962,
                         "BETA": 2.218},
              "WV_073": {"VC": 1362.081,
                         "ALPHA": 0.9991,
                         "BETA": 0.478},
              "IR_087": {"VC": 1149.069,
                         "ALPHA": 0.9996,
                         "BETA": 0.179},
              "IR_097": {"VC": 1034.343,
                         "ALPHA": 0.9999,
                         "BETA": 0.06},
              "IR_108": {"VC": 930.647,
                         "ALPHA": 0.9983,
                         "BETA": 0.625},
              "IR_120": {"VC": 839.66,
                         "ALPHA": 0.9988,
                         "BETA": 0.397},
              "IR_134": {"VC": 752.387,
                         "ALPHA": 0.9981,
                         "BETA": 0.578}}

# Meteosat 9
CALIB[322] = {"HRV": {"F": 79.0113},
              "VIS006": {"F": 65.2065},
              "VIS008": {"F": 73.1869},
              "IR_016": {"F": 61.9923},
              "IR_039": {"VC": 2568.832,
                         "ALPHA": 0.9954,
                         "BETA": 3.438},
              "WV_062": {"VC": 1600.548,
                         "ALPHA": 0.9963,
                         "BETA": 2.185},
              "WV_073": {"VC": 1360.330,
                         "ALPHA": 0.9991,
                         "BETA": 0.47},
              "IR_087": {"VC": 1148.620,
                         "ALPHA": 0.9996,
                         "BETA": 0.179},
              "IR_097": {"VC": 1035.289,
                         "ALPHA": 0.9999,
                         "BETA": 0.056},
              "IR_108": {"VC": 931.7,
                         "ALPHA": 0.9983,
                         "BETA": 0.64},
              "IR_120": {"VC": 836.445,
                         "ALPHA": 0.9988,
                         "BETA": 0.408},
              "IR_134": {"VC": 751.792,
                         "ALPHA": 0.9981,
                         "BETA": 0.561}}

# Meteosat 10
CALIB[323] = {"HRV": {"F": 78.9416},
              "VIS006": {"F": 65.5148},
              "VIS008": {"F": 73.1807},
              "IR_016": {"F": 62.0208},
              "IR_039": {"VC": 2547.771,
                         "ALPHA": 0.9915,
                         "BETA": 2.9002},
              "WV_062": {"VC": 1595.621,
                         "ALPHA": 0.9960,
                         "BETA": 2.0337},
              "WV_073": {"VC": 1360.337,
                         "ALPHA": 0.9991,
                         "BETA": 0.4340},
              "IR_087": {"VC": 1148.130,
                         "ALPHA": 0.9996,
                         "BETA": 0.1714},
              "IR_097": {"VC": 1034.715,
                         "ALPHA": 0.9999,
                         "BETA": 0.0527},
              "IR_108": {"VC": 929.842,
                         "ALPHA": 0.9983,
                         "BETA": 0.6084},
              "IR_120": {"VC": 838.659,
                         "ALPHA": 0.9988,
                         "BETA": 0.3882},
              "IR_134": {"VC": 750.653,
                         "ALPHA": 0.9982,
                         "BETA": 0.5390}}

# Meteosat 11
CALIB[324] = {"HRV": {"F": 79.0035},
              "VIS006": {"F": 65.2656},
              "VIS008": {"F": 73.1692},
              "IR_016": {"F": 61.9416},
              "IR_039": {"VC": 2555.280,
                         "ALPHA": 0.9916,
                         "BETA": 2.9438},
              "WV_062": {"VC": 1596.080,
                         "ALPHA": 0.9959,
                         "BETA": 2.0780},
              "WV_073": {"VC": 1361.748,
                         "ALPHA": 0.9990,
                         "BETA": 0.4929},
              "IR_087": {"VC": 1147.433,
                         "ALPHA": 0.9996,
                         "BETA": 0.1731},
              "IR_097": {"VC": 1034.851,
                         "ALPHA": 0.9998,
                         "BETA": 0.0597},
              "IR_108": {"VC": 931.122,
                         "ALPHA": 0.9983,
                         "BETA": 0.6256},
              "IR_120": {"VC": 839.113,
                         "ALPHA": 0.9988,
                         "BETA": 0.4002},
              "IR_134": {"VC": 748.585,
                         "ALPHA": 0.9981,
                         "BETA": 0.5635}}

# Calibration coefficients from Meirink, J.F., R.A. Roebeling and P. Stammes, 2013:
# Inter-calibration of polar imager solar channels using SEVIRI, Atm. Meas. Tech., 6,
# 2495-2508, doi:10.5194/amt-6-2495-2013
#
# The coeffients in the 2023 entry have been obtained from the webpage
# https://msgcpp.knmi.nl/solar-channel-calibration.html on 2023-10-11.
#
# The coefficients are stored in pairs of A, B (see function `get_meirink_slope`) where the
# units of A are µW m-2 sr-1 (cm-1)-1 and those of B are µW m-2 sr-1 (cm-1)-1 (86400 s)-1
#
# To obtain the slope for the calibration, one should use the routine get_seviri_meirink_slope

# Epoch for the MEIRINK re-calibration
MEIRINK_EPOCH = dt.datetime(2000, 1, 1)

MEIRINK_COEFS: dict[str, dict[int, dict[str, tuple[float, float]]]] = {}
MEIRINK_COEFS["2023"] = {}

# Meteosat-8

MEIRINK_COEFS["2023"][321] = {"VIS006": (24.346, 0.3739),
                              "VIS008": (30.989, 0.3111),
                              "IR_016": (22.869, 0.0065)
                              }

# Meteosat-9

MEIRINK_COEFS["2023"][322] = {"VIS006": (21.026, 0.2556),
                              "VIS008": (26.875, 0.1835),
                              "IR_016": (21.394, 0.0498)
                              }

# Meteosat-10

MEIRINK_COEFS["2023"][323] = {"VIS006": (19.829, 0.5856),
                              "VIS008": (25.284, 0.6787),
                              "IR_016": (23.066, -0.0286)
                              }

# Meteosat-11

MEIRINK_COEFS["2023"][324] = {"VIS006": (20.515, 0.3600),
                              "VIS008": (25.803, 0.4844),
                              "IR_016": (22.354, -0.0187)
                              }


def get_cds_time(days, msecs):
    """Compute timestamp given the days since epoch and milliseconds of the day.

    1958-01-01 00:00 is interpreted as fill value and will be replaced by NaT (Not a Time).

    Args:
        days (int, numpy.ndarray):
            Days since 1958-01-01
        msecs (int, numpy.ndarray):
            Milliseconds of the day

    Returns:
        numpy.datetime64: Timestamp(s)

    """
    if np.isscalar(days):
        days = np.array([days], dtype="int64")
        msecs = np.array([msecs], dtype="int64")

    # use nanosecond precision to silence warning from XArray
    nsecs = 1000000 * msecs.astype("timedelta64[ns]")
    time = np.datetime64("1958-01-01").astype("datetime64[ms]") + \
        days.astype("timedelta64[D]") + nsecs
    time[time == np.datetime64("1958-01-01 00:00")] = np.datetime64("NaT")

    if len(time) == 1:
        return time[0]
    return time


def add_scanline_acq_time(dataset, acq_time):
    """Add scanline acquisition time to the given dataset."""
    dataset.coords["acq_time"] = ("y", acq_time)
    dataset.coords["acq_time"].attrs[
        "long_name"] = "Mean scanline acquisition time"


def dec10216(inbuf):
    """Decode 10 bits data into 16 bits words.

    ::

        /*
         * pack 4 10-bit words in 5 bytes into 4 16-bit words
         *
         * 0       1       2       3       4       5
         * 01234567890123456789012345678901234567890
         * 0         1         2         3         4
         */
        ip = &in_buffer[i];
        op = &out_buffer[j];
        op[0] = ip[0]*4 + ip[1]/64;
        op[1] = (ip[1] & 0x3F)*16 + ip[2]/16;
        op[2] = (ip[2] & 0x0F)*64 + ip[3]/4;
        op[3] = (ip[3] & 0x03)*256 +ip[4];

    """
    arr10 = inbuf.astype(np.uint16)
    arr16_len = int(len(arr10) * 4 / 5)
    arr10_len = int((arr16_len * 5) / 4)
    arr10 = arr10[:arr10_len]  # adjust size

    # dask is slow with indexing
    arr10_0 = arr10[::5]
    arr10_1 = arr10[1::5]
    arr10_2 = arr10[2::5]
    arr10_3 = arr10[3::5]
    arr10_4 = arr10[4::5]

    arr16_0 = (arr10_0 << 2) + (arr10_1 >> 6)
    arr16_1 = ((arr10_1 & 63) << 4) + (arr10_2 >> 4)
    arr16_2 = ((arr10_2 & 15) << 6) + (arr10_3 >> 2)
    arr16_3 = ((arr10_3 & 3) << 8) + arr10_4
    arr16 = np.stack([arr16_0, arr16_1, arr16_2, arr16_3], axis=-1).ravel()

    return arr16


class MpefProductHeader(object):
    """MPEF product header class."""

    def get(self):
        """Return numpy record_array for MPEF product header."""
        record = [
            ("MPEF_File_Id", np.int16),
            ("MPEF_Header_Version", np.uint8),
            ("ManualDissAuthRequest", bool),
            ("ManualDisseminationAuth", bool),
            ("DisseminationAuth", bool),
            ("NominalTime", time_cds_short),
            ("ProductQuality", np.uint8),
            ("ProductCompleteness", np.uint8),
            ("ProductTimeliness", np.uint8),
            ("ProcessingInstanceId", np.int8),
            ("ImagesUsed", self.images_used, (4,)),
            ("BaseAlgorithmVersion",
             issue_revision),
            ("ProductAlgorithmVersion",
             issue_revision),
            ("InstanceServerName", "S2"),
            ("SpacecraftName", "S2"),
            ("Mission", "S3"),
            ("RectificationLongitude", "S5"),
            ("Encoding", "S1"),
            ("TerminationSpace", "S1"),
            ("EncodingVersion", np.uint16),
            ("Channel", np.uint8),
            ("ImageLocation", "S3"),
            ("GsicsCalMode", np.bool_),
            ("GsicsCalValidity", np.bool_),
            ("Padding", "S2"),
            ("OffsetToData", np.uint32),
            ("Padding2", "S9"),
            ("RepeatCycle", "S15"),
        ]

        return np.dtype(record).newbyteorder(">")

    @property
    def images_used(self):
        """Return structure for images_used."""
        record = [
            ("Padding1", "S2"),
            ("ExpectedImage", time_cds_short),
            ("ImageReceived", bool),
            ("Padding2", "S1"),
            ("UsedImageStart_Day", np.uint16),
            ("UsedImageStart_Millsec", np.uint32),
            ("Padding3", "S2"),
            ("UsedImageEnd_Day", np.uint16),
            ("UsedImageEndt_Millsec", np.uint32),
        ]

        return record


mpef_product_header = MpefProductHeader().get()



class IRCalibrationType(IntEnum):
    """IR calibration types.

    Corresponds to header entry "PlannedChanProcessing".
    """
    spectral_radiance = 1
    effective_radiance = 2
    not_processed = 0


class SEVIRICalibrationAlgorithm:
    """SEVIRI calibration algorithms."""

    def __init__(self, platform_id, scan_time):
        """Initialize the calibration algorithm."""
        self._platform_id = platform_id
        self._scan_time = scan_time

    def convert_to_radiance(self, data, gain, offset):
        """Calibrate to radiance."""
        data = data.where(data > 0)
        return (data * gain + offset).clip(0.0, None)

    def _erads2bt(self, data, channel_name):
        """Convert effective radiance to brightness temperature."""
        cal_info = CALIB[self._platform_id][channel_name]
        alpha = cal_info["ALPHA"]
        beta = cal_info["BETA"]
        wavenumber = CALIB[self._platform_id][channel_name]["VC"]

        return (self._tl15(data, wavenumber) - beta) / alpha

    def ir_calibrate(self, data: xr.DataArray, channel_name: str, cal_type: int) -> xr.DataArray:
        """Calibrate to brightness temperature."""
        if cal_type == IRCalibrationType.spectral_radiance:
            return self._srads2bt(data, channel_name)
        elif cal_type == IRCalibrationType.effective_radiance:
            return self._erads2bt(data, channel_name)
        elif cal_type == IRCalibrationType.not_processed:
            raise ValueError(f"No calibration coefficients available for {channel_name} "
                            "(PlannedChanProcessing is zero)")
        raise NotImplementedError(f"Unknown calibration type: {cal_type}. "
                                  f"Supported values are {list(IRCalibrationType)}.")

    def _srads2bt(self, data, channel_name):
        """Convert spectral radiance to brightness temperature."""
        a__, b__, c__ = BTFIT[channel_name]
        wavenumber = CALIB[self._platform_id][channel_name]["VC"]
        temp = self._tl15(data, wavenumber)

        return a__ * temp * temp + b__ * temp + c__

    def _tl15(self, data, wavenumber):
        """Compute the L15 temperature."""
        return ((C2 * wavenumber) /
                np.log((1.0 / data) * C1 * wavenumber ** 3 + 1.0))

    def vis_calibrate(self, data, solar_irradiance):
        """Calibrate to reflectance.

        This uses the method described in Conversion from radiances to
        reflectances for SEVIRI warm channels: https://user.eumetsat.int/s3/eup-strapi-media/pdf_msg_seviri_rad2refl_d6cd0c663f.pdf
        """
        reflectance = np.pi * data * 100.0 / solar_irradiance
        return utils.apply_earthsun_distance_correction(reflectance, self._scan_time)


CalibParams = namedtuple("CalibParams", ["mode", "internal_coefs", "external_coefs", "radiance_type"])
ScanParams = namedtuple("ScanParams", ["platform_id", "channel_name", "scan_time"])


class SEVIRICalibrationHandler:
    """Calibration handler for SEVIRI HRIT-, native- and netCDF-formats.

    Handles selection of calibration coefficients and calls the appropriate
    calibration algorithm.
    """

    def __init__(self, calib_params, scan_params):
        """Initialize the calibration handler."""
        self._calib_params = calib_params
        self._scan_params = scan_params
        self._algo = SEVIRICalibrationAlgorithm(
            platform_id=scan_params.platform_id,
            scan_time=scan_params.scan_time
        )
        self._check_calib_mode(calib_params.mode)

    def _check_calib_mode(self, calib_mode):
        valid_modes = ("NOMINAL", "GSICS", "MEIRINK-2023")
        if calib_mode not in valid_modes:
            raise ValueError(
                "Invalid calibration mode: {}. Choose one of {}".format(
                    calib_mode, valid_modes)
            )

    def calibrate(self, data, calibration):
        """Calibrate the given data."""
        if calibration == "counts":
            res = data
        elif calibration in ["radiance", "reflectance",
                             "brightness_temperature"]:
            coefs = self.get_coefs()
            res = self._algo.convert_to_radiance(
                data.astype(np.float32),
                np.float32(coefs["coefs"]["gain"]),
                np.float32(coefs["coefs"]["offset"])
            )
        else:
            raise ValueError(
                "Invalid calibration {} for channel {}".format(
                    calibration, self._scan_params.channel_name
                )
            )

        if calibration == "reflectance":
            solar_irradiance = CALIB[self._scan_params.platform_id][self._scan_params.channel_name]["F"]
            res = self._algo.vis_calibrate(res, solar_irradiance)
        elif calibration == "brightness_temperature":
            res = self._algo.ir_calibrate(
                res, self._scan_params.channel_name, self._calib_params.radiance_type
            )
        return res

    def get_coefs(self):
        """Get calibration coefficients."""
        picker = utils.CalibrationCoefficientPicker(self._calib_params.internal_coefs,
                                                    self._get_calib_wishlist(),
                                                    default="NOMINAL",
                                                    fallback="NOMINAL")
        return picker.get_coefs(self._scan_params.channel_name)

    def _get_calib_wishlist(self):
        ext_coefs = self._calib_params.external_coefs or {}
        wishlist = {
            ch: self._calib_params.mode for ch in CHANNEL_NAMES.values()
        }
        return wishlist | ext_coefs


def chebyshev(coefs: Sequence | np.ndarray, time: int | float, domain: Iterable):
    """Evaluate a Chebyshev Polynomial.

    Args:
        coefs: Coefficients defining the polynomial
        time: Time where to evaluate the polynomial
        domain: Domain (or time interval) for which the polynomial is defined: [left, right]
    Reference: Appendix A in the MSG Level 1.5 Image Data Format Description.

    """
    return Chebyshev(coefs, domain=domain)(time) - 0.5 * coefs[0]


def chebyshev_3d(coefs, time, domain):
    """Evaluate Chebyshev Polynomials for three dimensions (x, y, z).

    Expects the three coefficient sets to be defined in the same domain.

    Args:
        coefs: (x, y, z) coefficient sets.
        time: See :func:`chebyshev`
        domain: See :func:`chebyshev`

    Returns:
        Polynomials evaluated in (x, y, z) dimension.
    """
    x_coefs, y_coefs, z_coefs = coefs
    x = chebyshev(x_coefs, time, domain)
    y = chebyshev(y_coefs, time, domain)
    z = chebyshev(z_coefs, time, domain)
    return x, y, z


class NoValidOrbitParams(Exception):
    """Exception when validOrbitParameters are missing."""

    pass


class OrbitPolynomial:
    """Polynomial encoding the satellite position.

    Satellite position as a function of time is encoded in the coefficients
    of an 8th-order Chebyshev polynomial.
    """

    def __init__(self, coefs, start_time, end_time):
        """Initialize the polynomial."""
        self.coefs = coefs
        self.start_time = start_time
        self.end_time = end_time

    def evaluate(self, time):
        """Get satellite position in earth-centered cartesian coordinates.

        Args:
            time: Timestamp where to evaluate the polynomial

        Returns:
            Earth-centered cartesian coordinates (x, y, z) in meters
        """
        domain = [np.datetime64(self.start_time).astype("int64"),
                  np.datetime64(self.end_time).astype("int64")]
        time = np.datetime64(time).astype("int64")
        x, y, z = chebyshev_3d(self.coefs, time, domain)
        return x * 1000, y * 1000, z * 1000  # km -> m

    def __eq__(self, other):
        """Test equality of two orbit polynomials."""
        return (
            np.array_equal(self.coefs, np.array(other.coefs)) and
            self.start_time == other.start_time and
            self.end_time == other.end_time
        )


def get_satpos(orbit_polynomial, time, semi_major_axis, semi_minor_axis):
    """Get satellite position in geodetic coordinates.

    Args:
        orbit_polynomial: OrbitPolynomial instance
        time: Timestamp where to evaluate the polynomial
        semi_major_axis: Semi-major axis of the ellipsoid
        semi_minor_axis: Semi-minor axis of the ellipsoid

    Returns:
        Longitude [deg east], Latitude [deg north] and Altitude [m]
    """
    x, y, z = orbit_polynomial.evaluate(time)
    geocent = pyproj.CRS(
        proj="geocent", a=semi_major_axis, b=semi_minor_axis, units="m"
    )
    latlong = pyproj.CRS(
        proj="latlong", a=semi_major_axis, b=semi_minor_axis, units="m"
    )
    transformer = pyproj.Transformer.from_crs(geocent, latlong)
    lon, lat, alt = transformer.transform(x, y, z)
    return lon, lat, alt


class OrbitPolynomialFinder:
    """Find orbit polynomial for a given timestamp."""

    def __init__(self, orbit_polynomials):
        """Initialize with the given candidates.

        Args:
            orbit_polynomials: Dictionary of orbit polynomials as found in
                               SEVIRI L1B files:

                               .. code-block:: python

                                   {'X': x_polynomials,
                                    'Y': y_polynomials,
                                    'Z': z_polynomials,
                                    'StartTime': polynomials_valid_from,
                                    'EndTime': polynomials_valid_to}

        """
        self.orbit_polynomials = orbit_polynomials
        # Left/right boundaries of time intervals for which the polynomials are
        # valid.
        self.valid_from = orbit_polynomials["StartTime"][0, :].astype(
            "datetime64[us]")
        self.valid_to = orbit_polynomials["EndTime"][0, :].astype(
            "datetime64[us]")

    def get_orbit_polynomial(self, time, max_delta=6):
        """Get orbit polynomial valid for the given time.

        Orbit polynomials are only valid for certain time intervals. Find the
        polynomial, whose corresponding interval encloses the given timestamp.
        If there are multiple enclosing intervals, use the most recent one.
        If there is no enclosing interval, find the interval whose centre is
        closest to the given timestamp (but not more than ``max_delta`` hours
        apart).

        Why are there gaps between those intervals? Response from EUM:

        A manoeuvre is a discontinuity in the orbit parameters. The flight
        dynamic algorithms are not made to interpolate over the time-span of
        the manoeuvre; hence we have elements describing the orbit before a
        manoeuvre and a new set of elements describing the orbit after the
        manoeuvre. The flight dynamic products are created so that there is
        an intentional gap at the time of the manoeuvre. Also the two
        pre-manoeuvre elements may overlap. But the overlap is not of an
        issue as both sets of elements describe the same pre-manoeuvre orbit
        (with negligible variations).
        """
        time = np.datetime64(time)
        try:
            match = self._get_enclosing_interval(time)
        except ValueError:
            warnings.warn(
                "No orbit polynomial valid for {}. Using closest "
                "match.".format(time),
                stacklevel=2
            )
            match = self._get_closest_interval_within(time, max_delta)
        return OrbitPolynomial(
            coefs=(
                self.orbit_polynomials["X"][match],
                self.orbit_polynomials["Y"][match],
                self.orbit_polynomials["Z"][match]
            ),
            start_time=self.valid_from[match],
            end_time=self.valid_to[match]
        )

    def _get_enclosing_interval(self, time):
        """Find interval enclosing the given timestamp."""
        enclosing = np.where(
            np.logical_and(
                time >= self.valid_from,
                time < self.valid_to
            )
        )[0]
        most_recent = np.argmax(self.valid_from[enclosing])
        return enclosing[most_recent]

    def _get_closest_interval_within(self, time, threshold):
        """Find interval closest to the given timestamp within a given distance.

        Args:
            time: Timestamp of interest
            threshold: Maximum distance between timestamp and interval center

        Returns:
            Index of closest interval
        """
        closest_match, distance = self._get_closest_interval(time)
        threshold_diff = np.timedelta64(threshold, "h")
        if distance < threshold_diff:
            return closest_match
        raise NoValidOrbitParams(
            "Unable to find orbit coefficients valid for {} +/- {}"
            "hours".format(time, threshold)
        )

    def _get_closest_interval(self, time):
        """Find interval closest to the given timestamp.

        Returns:
            Index of closest interval, distance from its center
        """
        intervals_centre = self.valid_from + 0.5 * (
                self.valid_to - self.valid_from
        )
        diffs_us = (time - intervals_centre).astype("i8")
        closest_match = np.argmin(np.fabs(diffs_us))
        distance = abs(intervals_centre[closest_match] - time)
        return closest_match, distance


# def calculate_area_extent(center_point, north, east, south, west, we_offset, ns_offset, column_step, line_step):
def calculate_area_extent(area_dict):
    """Calculate the area extent seen by a geostationary satellite.

    Args:
        area_dict: A dictionary containing the required parameters
            center_point: Center point for the projection
            north: Northmost row number
            east: Eastmost column number
            west: Westmost column number
            south: Southmost row number
            column_step: Pixel resolution in meters in east-west direction
            line_step: Pixel resolution in meters in south-north direction
            [column_offset: Column offset, defaults to 0 if not given]
            [line_offset: Line offset, defaults to 0 if not given]

    Returns:
        tuple: An area extent for the scene defined by the lower left and
               upper right corners

    # For Earth model 2 and full disk VISIR, (center_point - west - 0.5 + we_offset) must be -1856.5 .
    # See MSG Level 1.5 Image Data Format Description Figure 7 - Alignment and numbering of the non-HRV pixels.
    """
    center_point = area_dict["center_point"]
    east = area_dict["east"]
    west = area_dict["west"]
    south = area_dict["south"]
    north = area_dict["north"]
    column_step = area_dict["column_step"]
    line_step = area_dict["line_step"]
    column_offset = area_dict.get("column_offset", 0)
    line_offset = area_dict.get("line_offset", 0)

    ll_c = (center_point - east + 0.5 + column_offset) * column_step
    ll_l = (north - center_point + 0.5 + line_offset) * line_step
    ur_c = (center_point - west - 0.5 + column_offset) * column_step
    ur_l = (south - center_point - 0.5 + line_offset) * line_step

    return (ll_c, ll_l, ur_c, ur_l)


def create_coef_dict(nominal_coefs, gsics_coefs=None, meirink_coefs=None):
    """Create coefficient dictionary expected by calibration class."""
    coefs = nominal_coefs.get_coefs()
    if gsics_coefs:
        coefs.update(gsics_coefs.get_coefs())
    if meirink_coefs:
        coefs.update(meirink_coefs.get_coefs(nominal_coefs.offset))
    return coefs


class NominalCoefficients:
    """Nominal calibration coefficients."""
    def __init__(self, channel_name, gain, offset):
        """Initialize coefficients."""
        self.channel_name = channel_name
        self.gain = gain
        self.offset = offset

    def get_coefs(self):
        """Get coefficient dictionary."""
        return {
            "NOMINAL": {
                self.channel_name: {
                    "gain": self.gain,
                    "offset": self.offset
                }
            }
        }


class GsicsCoefficients:
    """GSICS calibration coefficients."""
    def __init__(self, channel_name, gain, offset):
        """Initialize coefficients."""
        self.channel_name = channel_name
        self.gain = gain
        self.offset = offset

    def get_coefs(self):
        """Get coefficient dictionary."""
        coefs = {"GSICS": {}}
        if self._is_available():
            coefs["GSICS"][self.channel_name] = {
                "gain": self.gain,
                "offset": self.offset * self.gain
            }
        return coefs

    def _is_available(self):
        # If no GSICS coefficients are available they are set to zero in
        # the file.
        return self.gain != 0 and self.offset != 0


class MeirinkCoefficients:
    """Re-calibration of the SEVIRI visible channels slope (see Meirink 2013)."""

    def __init__(self, platform_id, channel_name, scan_time):
        """Initialize coefficients."""
        self.platform_id = platform_id
        self.channel_name = channel_name
        self.scan_time = scan_time

    def get_coefs(self, offset):
        """Get coefficient dictionary.

        Args:
            offset: Nominal calibration offset.
        """
        gain = self._get_gain()
        return self._combine_gain_and_offset(gain, offset)

    def _get_gain(self):
        res = {}
        for version, coefs in MEIRINK_COEFS.items():
            gain = self._get_gain_single_channel(coefs)
            if gain:
                res[f"MEIRINK-{version}"] = gain
        return res

    def _get_gain_single_channel(self, coefs):
        try:
            coefs_ch = coefs[self.platform_id][self.channel_name]
            return self.get_slope(coefs_ch, self.scan_time)
        except KeyError:
            return None

    @staticmethod
    def get_slope(coefs_single_channel, acquisition_time):
        """Compute the slope for the visible channel calibration according to Meirink 2013.

        S = A + B * 1.e-3* Day

        S is here in µW m-2 sr-1 (cm-1)-1

        EUMETSAT calibration is given in mW m-2 sr-1 (cm-1)-1, so an extra factor of 1/1000 must
        be applied.
        """
        A = coefs_single_channel[0]
        B = coefs_single_channel[1]
        delta_t = (acquisition_time - MEIRINK_EPOCH).total_seconds()
        S = A + B * delta_t / (3600*24) / 1000.
        return S/1000

    def _combine_gain_and_offset(self, gain, offset):
        return {
            calib_mode: {
                self.channel_name: {"gain": gain_, "offset": offset}
            }
                for calib_mode, gain_ in gain.items()
        }

def get_padding_area(shape, dtype):
    """Create a padding area filled with no data."""
    if np.issubdtype(dtype, np.floating):
        init_value = np.nan
    else:
        init_value = 0

    padding_area = da.full(shape, init_value, dtype=dtype, chunks=CHUNK_SIZE)

    return padding_area


def pad_data_horizontally(data, final_size, east_bound, west_bound):
    """Pad the data given east and west bounds and the desired size."""
    nlines = final_size[0]
    if west_bound - east_bound != data.shape[1] - 1:
        raise IndexError("East and west bounds do not match data shape")

    padding_east = get_padding_area((nlines, east_bound - 1), data.dtype)
    padding_west = get_padding_area((nlines, (final_size[1] - west_bound)), data.dtype)

    return np.hstack((padding_east, data, padding_west))


def pad_data_vertically(data, final_size, south_bound, north_bound):
    """Pad the data given south and north bounds and the desired size."""
    ncols = final_size[1]
    if north_bound - south_bound != data.shape[0] - 1:
        raise IndexError("South and north bounds do not match data shape")

    padding_south = get_padding_area((south_bound - 1, ncols), data.dtype)
    padding_north = get_padding_area(((final_size[0] - north_bound), ncols), data.dtype)

    return np.vstack((padding_south, data, padding_north))


def _create_bad_quality_lines_mask(line_validity, line_geometric_quality, line_radiometric_quality):
    """Create bad quality scan lines mask.

    For details on quality flags see `MSG Level 1.5 Image Data Format Description`_
    page 109.

    Args:
        line_validity (numpy.ndarray):
            Quality flags with shape (nlines,).
        line_geometric_quality (numpy.ndarray):
            Quality flags with shape (nlines,).
        line_radiometric_quality (numpy.ndarray):
            Quality flags with shape (nlines,).

    Returns:
        numpy.ndarray: Indicating if the scan line is bad.
    """
    # Based on missing (2) or corrupted (3) data
    line_mask = line_validity >= 2
    line_mask &= line_validity <= 3
    # Do not use (4)
    line_mask &= line_radiometric_quality == 4
    line_mask &= line_geometric_quality == 4
    return line_mask


def mask_bad_quality(data, line_validity, line_geometric_quality, line_radiometric_quality):
    """Mask scan lines with bad quality.

    Args:
        data (xarray.DataArray):
            Channel data
        line_validity (numpy.ndarray):
            Quality flags with shape (nlines,).
        line_geometric_quality (numpy.ndarray):
            Quality flags with shape (nlines,).
        line_radiometric_quality (numpy.ndarray):
            Quality flags with shape (nlines,).

    Returns:
        xarray.DataArray: data with lines flagged as bad converted to np.nan.
    """
    line_mask = _create_bad_quality_lines_mask(line_validity, line_geometric_quality, line_radiometric_quality)
    line_mask = line_mask[:, np.newaxis]
    data = data.where(~line_mask, np.nan).astype(np.float32)
    return data


def round_nom_time(date, time_delta):
    """Round a datetime object to a multiple of a timedelta.

    date : datetime.datetime object, default now.
    time_delta : timedelta object, we round to a multiple of this, default 1 minute.
    adapted for SEVIRI from:
    https://stackoverflow.com/questions/3463930/how-to-round-the-minute-of-a-datetime-object-python
    """
    seconds = (date - date.min).seconds
    round_to = time_delta.total_seconds()

    rounding = (seconds + round_to / 2) // round_to * round_to

    return date + dt.timedelta(0, rounding - seconds, - date.microsecond)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utilities for dealing with ancillary variables."""

from .dataid import DataID, default_id_keys_config


def dataset_walker(datasets):
    """Walk through *datasets* and their ancillary data.

    Yields datasets and their parent.
    """
    for dataset in datasets:
        yield dataset, None
        for anc_ds in dataset.attrs.get("ancillary_variables", []):
            try:
                anc_ds.attrs
                yield anc_ds, dataset
            except AttributeError:
                continue


def replace_anc(dataset, parent_dataset):
    """Replace *dataset* the *parent_dataset*'s `ancillary_variables` field."""
    if parent_dataset is None:
        return
    id_keys = parent_dataset.attrs.get(
            "_satpy_id_keys",
            dataset.attrs.get(
                "_satpy_id_keys",
                default_id_keys_config))
    current_dataid = DataID(id_keys, **dataset.attrs)
    for idx, ds in enumerate(parent_dataset.attrs["ancillary_variables"]):
        if current_dataid == DataID(id_keys, **ds.attrs):
            parent_dataset.attrs["ancillary_variables"][idx] = dataset
            return



# Copyright (c) 2024-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Script to create image testing references.

Script to create reference images for the automated image testing system.

The input data directory must follow the data structure from the
image-comparison-tests repository with satellite_data/<satellite-name>.

This script is a work in progress and expected to change significantly.

DO NOT USE FOR OPERATIONAL PRODUCTION!
"""

import argparse
import os
import pathlib

import hdf5plugin  # noqa: F401

from satpy import Scene


def generate_images(props):
    """Generate reference images for testing purposes.

    Args:
        props (namespace): Object with attributes corresponding to command line
        arguments as defined by :func:get_parser.
    """
    filenames = (props.basedir / "satellite_data" / props.satellite /
                 props.case).glob("*")

    if "," in props.reader:
        reader = props.reader.split(",")
        resampler = "nearest"  # use nearest when combining with cloud mask
    else:
        reader = props.reader
        resampler = "gradient_search"
    scn = Scene(reader=reader, filenames=filenames)

    scn.load(props.composites)
    if props.area == "native":
        ls = scn.resample(resampler="native")
    elif props.area is not None:
        ls = scn.resample(props.area, resampler=resampler)
    else:
        ls = scn

    from dask.diagnostics import ProgressBar
    with ProgressBar():
        ls.save_datasets(
                writer="simple_image",
                filename=os.fspath(
                    props.basedir / "reference_images" /
                    "satpy-reference-image-{platform_name}-{sensor}-"
                    "{start_time:%Y%m%d%H%M}-{area.area_id}-{name}.png"))

def get_parser():
    """Return argument parser."""
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
            "satellite", action="store", type=str,
            help="Satellite name.")

    parser.add_argument(
            "reader", action="store", type=str,
            help="Reader name.  Multiple readers (if needed) can be comma-seperated.")

    parser.add_argument(
            "case", help="case to generate", type=str)

    parser.add_argument(
            "-b", "--basedir", action="store", type=pathlib.Path,
            default=pathlib.Path("."),
            help="Base directory for reference data. "
                 "This must contain a subdirectories satellite_data and "
                 "reference_images.  The directory satellite_data must contain "
                 "input data in a subdirectory for the satellite and case. Output images "
                 "will be written to the subdirectory reference_images.")

    parser.add_argument(
            "-c", "--composites", nargs="+", help="composites to generate",
            type=str, default=["ash", "airmass"])

    parser.add_argument(
            "-a", "--area", action="store",
            default=None,
            help="Area name, or 'native' (native resampling)")

    return parser

def main():
    """Main function."""
    parsed = get_parser().parse_args()

    generate_images(parsed)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright (c) 2016-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helpers for reading hdf5-based files."""

from __future__ import annotations

from typing import Any

from satpy.utils import _import_and_warn_new_location


def __getattr__(name: str) -> Any:
    new_module = "satpy.readers.core.hdf5"

    return _import_and_warn_new_location(new_module, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Satpy developers
#
# This file is part of Satpy.
#
# Satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the satpy.readers.tropomi_l2 module."""

import datetime as dt
import os
import unittest
from unittest import mock

import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler

DEFAULT_FILE_DTYPE = np.float32
DEFAULT_FILE_SHAPE = (9001, 18000)
DEFAULT_LAT = np.linspace(-90, 90, DEFAULT_FILE_SHAPE[0], dtype=DEFAULT_FILE_DTYPE)
DEFAULT_LON = np.linspace(-180, 180, DEFAULT_FILE_SHAPE[1], dtype=DEFAULT_FILE_DTYPE)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
file_content_units = dict()


class FakeNetCDF4FileHandlerMimic(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        from xarray import DataArray
        dt_s = filename_info.get("start_time", dt.datetime(2019, 6, 19, 13, 0))
        dt_e = filename_info.get("end_time", dt.datetime(2019, 6, 19, 13, 0))

        if filetype_info["file_type"] == "mimicTPW2_comp":
            file_content = {
                "/attr/start_time": dt_s.strftime("%Y%m%d.%H%M%S"),
                "/attr/end_time": dt_e.strftime("%Y%m%d.%H%M%S"),
                "/attr/platform_shortname": "aggregated microwave",
                "/attr/sensor": "mimic",
            }
            file_content["latArr"] = DEFAULT_LAT
            file_content["latArr/shape"] = (DEFAULT_FILE_SHAPE[0],)
            file_content["latArr/attr/units"] = "degress_north"

            file_content["lonArr"] = DEFAULT_LON
            file_content["lonArr/shape"] = (DEFAULT_FILE_SHAPE[1],)
            file_content["lonArr/attr/units"] = "degrees_east"

            file_content["tpwGrid"] = DEFAULT_FILE_DATA
            file_content["tpwGrid/shape"] = DEFAULT_FILE_SHAPE
            file_content_units["tpwGrid"] = "mm"

            file_content["/dimension/lat"] = DEFAULT_FILE_SHAPE[0]
            file_content["/dimension/lon"] = DEFAULT_FILE_SHAPE[1]

            # convert to xarrays
            for key, val in file_content.items():
                if key == "lonArr" or key == "latArr":
                    file_content[key] = DataArray(val)
                elif isinstance(val, np.ndarray):
                    if val.ndim > 1:
                        file_content[key] = DataArray(val, dims=("y", "x"), attrs={"units": file_content_units[key]})
                    else:
                        file_content[key] = DataArray(val)
        else:
            msg = "Wrong Test Reader for file_type {}".format(filetype_info["file_type"])
            raise AssertionError(msg)

        return file_content


class TestMimicTPW2Reader(unittest.TestCase):
    """Test Mimic Reader."""

    yaml_file = "mimicTPW2_comp.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.mimic_TPW2_nc import MimicTPW2FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(MimicTPW2FileHandler, "__bases__", (FakeNetCDF4FileHandlerMimic,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic initialization of this reader."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "comp20190619.130000.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_load_mimic(self):
        """Load Mimic data."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.mimic_TPW2_nc.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "comp20190619.130000.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(["tpwGrid"])
        assert len(ds) == 1
        for d in ds.values():
            assert d.attrs["platform_shortname"] == "aggregated microwave"
            assert d.attrs["sensor"] == "mimic"
            assert "area" in d.attrs
            assert "units" in d.attrs
            assert d.attrs["area"] is not None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit testing for the ABI enhancement functions."""

import unittest

import dask.array as da
import numpy as np
import xarray as xr


class TestABIEnhancement(unittest.TestCase):
    """Test the ABI enhancement functions."""

    def setUp(self):
        """Create fake data for the tests."""
        data = da.linspace(0, 1, 16).reshape((4, 4))
        self.da = xr.DataArray(data, dims=("y", "x"), attrs={"test": "test"})

    def test_cimss_true_color_contrast(self):
        """Test the cimss_true_color_contrast enhancement."""
        from trollimage.xrimage import XRImage

        from satpy.enhancements.abi import cimss_true_color_contrast

        expected = np.array([[
            [0., 0., 0.05261956, 0.13396146],
            [0.21530335, 0.29664525, 0.37798715, 0.45932905],
            [0.54067095, 0.62201285, 0.70335475, 0.78469665],
            [0.86603854, 0.94738044, 1., 1.],
            ]])
        img = XRImage(self.da)
        cimss_true_color_contrast(img)
        np.testing.assert_almost_equal(img.data.compute(), expected)



# Copyright (c) 2015-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Dataset identifying objects."""

import logging
import numbers
from collections import namedtuple
from contextlib import suppress
from copy import copy, deepcopy
from enum import Enum, IntEnum
from typing import NoReturn

import numpy as np

logger = logging.getLogger(__name__)


def get_keys_from_config(common_id_keys, config):
    """Gather keys for a new DataID from the ones available in configured dataset."""
    id_keys = {}
    for key, val in common_id_keys.items():
        if key in config:
            id_keys[key] = val
        elif val is not None and (val.get("required") is True or val.get("default") is not None):
            id_keys[key] = val
    if not id_keys:
        raise ValueError("Metadata does not contain enough information to create a DataID.")
    return id_keys


class ValueList(IntEnum):
    """A static value list.

    This class is meant to be used for dynamically created Enums. Due to this
    it should not be used as a normal Enum class or there may be some
    unexpected behavior. For example, this class contains custom pickling and
    unpickling handling that may break in subclasses.

    """

    @classmethod
    def convert(cls, value):
        """Convert value to an instance of this class."""
        try:
            return cls[value]
        except KeyError:
            raise ValueError("{} invalid value for {}".format(value, cls))

    @classmethod
    def _unpickle(cls, enum_name, enum_members, enum_member):
        """Create dynamic class that was previously pickled.

        See :meth:`ValueList.__reduce_ex__` for implementation details.

        """
        enum_cls = cls(enum_name, enum_members)
        return enum_cls[enum_member]

    def __reduce_ex__(self, proto):
        """Reduce the object for pickling."""
        return (ValueList._unpickle,
                (self.__class__.__name__, list(self.__class__.__members__.keys()), self.name))

    def __eq__(self, other):
        """Check equality."""
        return self.name == other

    def __ne__(self, other):
        """Check non-equality."""
        return self.name != other

    def __hash__(self):
        """Hash the object."""
        return hash(self.name)

    def __repr__(self):
        """Represent the values."""
        return "<" + str(self) + ">"


wlklass = namedtuple("WavelengthRange", "min central max unit", defaults=("µm",))  # type: ignore


class WavelengthRange(wlklass):
    """A named tuple for wavelength ranges.

    The elements of the range are min, central and max values, and optionally a unit
    (defaults to µm). No clever unit conversion is done here, it's just used for checking
    that two ranges are comparable.
    """

    def __eq__(self, other: object) -> bool:
        """Return if two wavelengths are equal.

        Args:
            other: (min wl, nominal wl, max wl) or scalar wl

        Return:
            True if other is a scalar and min <= other <= max, or if other is
            a tuple equal to self, False otherwise.

        """
        if other is None:
            return False
        if isinstance(other, numbers.Number):
            return other in self
        if isinstance(other, (tuple, list)) and len(other) == 3:
            return self[:3] == other
        return super().__eq__(other)

    def __ne__(self, other):
        """Return the opposite of `__eq__`."""
        return not self == other

    def __lt__(self, other):
        """Compare to another wavelength."""
        if other is None:
            return False
        return super().__lt__(other)

    def __gt__(self, other):
        """Compare to another wavelength."""
        if other is None:
            return True
        return super().__gt__(other)

    def __hash__(self):
        """Hash this tuple."""
        return tuple.__hash__(self)

    def __str__(self):
        """Format for print out."""
        return "{0.central} {0.unit} ({0.min}-{0.max} {0.unit})".format(self)

    def __contains__(self, other):
        """Check if this range contains *other*."""
        if other is None:
            return False
        if isinstance(other, numbers.Number):
            return self.min <= other <= self.max
        with suppress(AttributeError):
            if self.unit != other.unit:
                raise NotImplementedError("Can't compare wavelength ranges with different units.")
            return self.min <= other.min and self.max >= other.max
        return False

    def distance(self, value):
        """Get the distance from value."""
        if self == value:
            try:
                return abs(value.central - self.central)
            except AttributeError:
                if isinstance(value, (tuple, list)):
                    return abs(value[1] - self.central)
                return abs(value - self.central)
        else:
            return np.inf

    @classmethod
    def convert(cls, wl):
        """Convert `wl` to this type if possible."""
        if isinstance(wl, (tuple, list)):
            return cls(*wl)
        return wl

    def to_cf(self):
        """Serialize for cf export."""
        return str(self)

    @classmethod
    def from_cf(cls, blob):
        """Return a WavelengthRange from a cf blob."""
        try:
            obj = cls._read_cf_from_string_export(blob)
        except TypeError:
            obj = cls._read_cf_from_string_list(blob)
        return obj

    @classmethod
    def _read_cf_from_string_export(cls, blob):
        """Read blob as a string created by `to_cf`."""
        pattern = "{central:f} {unit:s} ({min:f}-{max:f} {unit2:s})"
        from trollsift import Parser
        parser = Parser(pattern)
        res_dict = parser.parse(blob)
        res_dict.pop("unit2")
        obj = cls(**res_dict)
        return obj

    @classmethod
    def _read_cf_from_string_list(cls, blob):
        """Read blob as a list of strings (legacy formatting)."""
        min_wl, central_wl, max_wl, unit = blob
        obj = cls(float(min_wl), float(central_wl), float(max_wl), unit)
        return obj


class ModifierTuple(tuple):
    """A tuple holder for modifiers."""

    @classmethod
    def convert(cls, modifiers):
        """Convert `modifiers` to this type if possible."""
        if modifiers is None:
            return None
        if not isinstance(modifiers, (cls, tuple, list)):
            raise TypeError("'DataID' modifiers must be a tuple or None, "
                            "not {}".format(type(modifiers)))
        return cls(modifiers)

    def __eq__(self, other):
        """Check equality."""
        if isinstance(other, list):
            other = tuple(other)
        return super().__eq__(other)

    def __ne__(self, other):
        """Check non-equality."""
        if isinstance(other, list):
            other = tuple(other)
        return super().__ne__(other)

    def __hash__(self):
        """Hash this tuple."""
        return tuple.__hash__(self)


#: Default ID keys DataArrays.
default_id_keys_config = {"name": {
                              "required": True,
                          },
                          "wavelength": {
                              "type": WavelengthRange,
                          },
                          "resolution": {
                              "transitive": False,
                              },
                          "calibration": {
                              "enum": [
                                  "reflectance",
                                  "brightness_temperature",
                                  "radiance",
                                  "radiance_wavenumber",
                                  "counts"
                                  ],
                              "transitive": True,
                          },
                          "modifiers": {
                              "default": ModifierTuple(),
                              "type": ModifierTuple,
                          },
                          }

#: Default ID keys for coordinate DataArrays.
default_co_keys_config = {"name": {
                              "required": True,
                          },
                          "resolution": {
                              "transitive": True,
                          }
                          }

#: Minimal ID keys for DataArrays, for example composites.
minimal_default_keys_config = {"name": {
                                  "required": True,
                              },
                               "resolution": {
                                   "transitive": True,
                               }
                              }


class DataID(dict):
    """Identifier for all `DataArray` objects.

    DataID is a dict that holds identifying and classifying
    information about a DataArray.
    """

    def __init__(self, id_keys, **keyval_dict):
        """Init the DataID.

        The *id_keys* dictionary has to be formed as described in :doc:`../dev_guide/satpy_internals`.
        The other keyword arguments are values to be assigned to the keys. Note that
        `None` isn't a valid value and will simply be ignored.
        """
        self._hash = None
        self._orig_id_keys = id_keys
        self._id_keys = self.fix_id_keys(id_keys or {})
        if keyval_dict:
            curated = self.convert_dict(keyval_dict)
        else:
            curated = {}
        super(DataID, self).__init__(curated)

    @staticmethod
    def fix_id_keys(id_keys):
        """Flesh out enums in the id keys as gotten from a config."""
        new_id_keys = id_keys.copy()
        for key, val in id_keys.items():
            if not val:
                continue
            if "enum" in val and "type" in val:
                raise ValueError("Cannot have both type and enum for the same id key.")
            new_val = copy(val)
            if "enum" in val:
                new_val["type"] = ValueList(key, " ".join(new_val.pop("enum")))
            new_id_keys[key] = new_val
        return new_id_keys

    def convert_dict(self, keyvals):
        """Convert a dictionary's values to the types defined in this object's id_keys."""
        curated = {}
        if not keyvals:
            return curated
        for key, val in self._id_keys.items():
            if val is None:
                val = {}
            if key in keyvals or val.get("default") is not None or val.get("required"):
                curated_val = keyvals.get(key, val.get("default"))
                if "required" in val and curated_val is None:
                    raise ValueError("Required field {} missing.".format(key))
                if "type" in val:
                    curated[key] = val["type"].convert(curated_val)
                elif curated_val is not None:
                    curated[key] = curated_val

        return curated

    @classmethod
    def _unpickle(cls, id_keys, keyval):
        """Create a new instance of the DataID after pickling."""
        return cls(id_keys, **keyval)

    def __reduce__(self):
        """Reduce the object for pickling."""
        return (self._unpickle, (self._orig_id_keys, self.to_dict()))

    def from_dict(self, keyvals):
        """Create a DataID from a dictionary."""
        return self.__class__(self._id_keys, **keyvals)

    @classmethod
    def from_dataarray(cls, array, default_keys=minimal_default_keys_config):
        """Get the DataID using the dataarray attributes."""
        if "_satpy_id" in array.attrs:
            return array.attrs["_satpy_id"]
        return cls.new_id_from_dataarray(array, default_keys)

    @classmethod
    def new_id_from_dataarray(cls, array, default_keys=minimal_default_keys_config):
        """Create a new DataID from a dataarray's attributes."""
        try:
            id_keys = array.attrs["_satpy_id"].id_keys
        except KeyError:
            id_keys = array.attrs.get("_satpy_id_keys", default_keys)
        return cls(id_keys, **array.attrs)

    @property
    def id_keys(self):
        """Get the id_keys."""
        return deepcopy(self._id_keys)

    def create_filter_query_without_required_fields(self, query):
        """Remove the required fields from *query*."""
        try:
            new_query = query.to_dict()
        except AttributeError:
            new_query = query.copy()
        for key, val in self._id_keys.items():
            if val and (val.get("transitive") is not True):
                new_query.pop(key, None)
        return DataQuery.from_dict(new_query)

    def _asdict(self):
        return dict(self.items())

    def to_dict(self):
        """Convert the ID to a dict."""
        res_dict = dict()
        for key, value in self._asdict().items():
            if isinstance(value, Enum):
                res_dict[key] = value.name
            else:
                res_dict[key] = value
        return res_dict

    def __deepcopy__(self, memo=None):
        """Copy this object.

        Returns self as it's immutable.
        """
        return self

    def __copy__(self):
        """Copy this object.

        Returns self as it's immutable.
        """
        return self

    def __repr__(self):
        """Represent the id."""
        items = ("{}={}".format(key, repr(val)) for key, val in self.items())
        return self.__class__.__name__ + "(" + ", ".join(items) + ")"

    def _replace(self, **kwargs):
        """Make a new instance with replaced items."""
        info = dict(self.items())
        info.update(kwargs)
        return self.from_dict(info)

    def __hash__(self):
        """Hash the object."""
        if self._hash is None:
            self._hash = hash(tuple(sorted(self.items())))
        return self._hash

    def _immutable(self, *args, **kws) -> NoReturn:
        """Raise and error."""
        raise TypeError("Cannot change a DataID")

    def __lt__(self, other):
        """Check lesser than."""
        list_self, list_other = [], []
        for key in self._id_keys:
            if key not in self and key not in other:
                continue
            elif key in self and key in other:
                list_self.append(self[key])
                list_other.append(other[key])
            elif key in self:
                val = self[key]
                list_self.append(val)
                list_other.append(_generalize_value_for_comparison(val))
            elif key in other:
                val = other[key]
                list_other.append(val)
                list_self.append(_generalize_value_for_comparison(val))
        return tuple(list_self) < tuple(list_other)

    __setitem__ = _immutable
    __delitem__ = _immutable
    pop = _immutable  # type: ignore
    popitem = _immutable
    clear = _immutable
    update = _immutable  # type: ignore
    setdefault = _immutable  # type: ignore

    def _find_modifiers_key(self):
        for key, val in self.items():
            if isinstance(val, ModifierTuple):
                return key
        raise KeyError

    def create_less_modified_query(self):
        """Create a query with one less modifier."""
        new_dict = self.to_dict()
        new_dict["modifiers"] = tuple(new_dict["modifiers"][:-1])
        return DataQuery.from_dict(new_dict)

    def is_modified(self):
        """Check if this is modified."""
        try:
            key = self._find_modifiers_key()
        except KeyError:
            return False
        return bool(self[key])


def _generalize_value_for_comparison(val):
    """Get a generalize value for comparisons."""
    if isinstance(val, numbers.Number):
        return 0
    if isinstance(val, str):
        return ""
    if isinstance(val, tuple):
        return tuple()

    raise NotImplementedError("Don't know how to generalize " + str(type(val)))


class DataQuery:
    """The data query object.

    A DataQuery can be used in Satpy to query for a Dataset. This way
    a fully qualified DataID can be found even if some DataID
    elements are unknown. In this case a `*` signifies something that is
    unknown or not applicable to the requested Dataset.
    """

    def __init__(self, **kwargs):
        """Initialize the query."""
        self._dict = kwargs.copy()
        self._fields = tuple(self._dict.keys())
        self._values = tuple(self._dict.values())

    def __getitem__(self, key):
        """Get an item."""
        return self._dict[key]

    def __eq__(self, other):
        """Compare the DataQuerys.

        A DataQuery is considered equal to another DataQuery or DataID
        if they have common keys that have equal values.
        """
        sdict = self._asdict()
        try:
            odict = other._asdict()
        except AttributeError:
            return False
        common_keys = False
        for key, val in sdict.items():
            if key in odict:
                common_keys = True
                if odict[key] != val and val is not None:
                    return False
        return common_keys

    def __hash__(self):
        """Hash."""
        fields = []
        values = []
        for field, value in sorted(self._dict.items()):
            if value != "*":
                fields.append(field)
                if isinstance(value, (list, set)):
                    value = tuple(value)
                values.append(value)
        return hash(tuple(zip(fields, values)))

    def get(self, key, default=None):
        """Get an item."""
        return self._dict.get(key, default)

    @classmethod
    def from_dict(cls, the_dict):
        """Convert a dict to an ID."""
        return cls(**the_dict)

    def items(self):
        """Get the items of this query."""
        return self._dict.items()

    def _asdict(self):
        return self._dict.copy()

    def to_dict(self, trim=True):
        """Convert the ID to a dict."""
        if trim:
            return self._to_trimmed_dict()
        else:
            return self._asdict()

    def _to_trimmed_dict(self):
        return {key: val for key, val in self._dict.items()
                if val != "*"}

    def __repr__(self):
        """Represent the query."""
        items = ("{}={}".format(key, repr(val)) for key, val in zip(self._fields, self._values))
        return self.__class__.__name__ + "(" + ", ".join(items) + ")"

    def filter_dataids(self, dataid_container):
        """Filter DataIDs based on this query."""
        keys = list(filter(self._match_dataid, dataid_container))

        return keys

    def _match_dataid(self, dataid):
        """Match the dataid with the current query."""
        if self._shares_required_keys(dataid):
            keys_to_check = set(dataid.keys()) & set(self._fields)
        else:
            keys_to_check = set(dataid._id_keys.keys()) & set(self._fields)
        if not keys_to_check:
            return False
        return all(self._match_query_value(key, dataid.get(key)) for key in keys_to_check)

    def _shares_required_keys(self, dataid):
        """Check if dataid shares required keys with the current query."""
        for key, val in dataid._id_keys.items():
            try:
                if val.get("required", False):
                    if key in self._fields:
                        return True
            except AttributeError:
                continue
        return False

    def _match_query_value(self, key, id_val):
        val = self._dict[key]
        if val == "*":
            return True
        if isinstance(id_val, tuple) and isinstance(val, (tuple, list)):
            return tuple(val) == id_val
        if not isinstance(val, list):
            val = [val]
        return id_val in val

    def sort_dataids_with_preference(self, all_ids, preference):
        """Sort `all_ids` given a sorting `preference` (DataQuery or None)."""
        try:
            res = preference.to_dict()
        except AttributeError:
            res = dict()
        res.update(self.to_dict())
        optimistic_query = DataQuery.from_dict(res)
        sorted_ids, distances = optimistic_query.sort_dataids(all_ids)
        if distances[0] == np.inf:  # nothing matches the optimistic query
            sorted_ids, distances = self.sort_dataids(all_ids)
        return sorted_ids, distances

    def sort_dataids(self, dataids):
        """Sort the DataIDs based on this query.

        Returns the sorted dataids and the list of distances.

        The sorting is performed based on the types of the keys to search on
        (as they are defined in the DataIDs from `dataids`).
        If that type defines a `distance` method, then it is used to find how
        'far' the DataID is from the current query.
        If the type is a number, a simple subtraction is performed.
        For other types, the distance is 0 if the values are identical, np.inf
        otherwise.

        For example, with the default DataID, we use the following criteria:

        1. Central wavelength is nearest to the `key` wavelength if
           specified.
        2. Least modified dataset if `modifiers` is `None` in `key`.
           Otherwise, the modifiers are ignored.
        3. Highest calibration if `calibration` is `None` in `key`.
           Calibration priority is the order of the calibration list defined as
           reflectance, brightness temperature, radiance counts if not overridden in the
           reader configuration.
        4. Best resolution (smallest number) if `resolution` is `None`
           in `key`. Otherwise, the resolution is ignored.

        """
        distances = []
        sorted_dataids = []
        big_distance = 100000
        keys = set(self._dict.keys())
        for dataid in dataids:
            keys |= set(dataid.keys())
        for dataid in sorted(dataids):
            sorted_dataids.append(dataid)
            distance = 0
            for key in keys:
                if distance == np.inf:
                    break
                val = self._dict.get(key, "*")
                if val == "*":
                    distance = self._add_absolute_distance(dataid, key, distance)
                else:
                    try:
                        dataid_val = dataid[key]
                    except KeyError:
                        distance += big_distance
                        continue
                    distance = self._add_distance_from_query(dataid_val, val, distance)
            distances.append(distance)
        distances, dataids = zip(*sorted(zip(distances, sorted_dataids)))
        return dataids, distances

    @staticmethod
    def _add_absolute_distance(dataid, key, distance):
        try:
            # for enums
            distance += dataid.get(key).value
        except AttributeError:
            if isinstance(dataid.get(key), numbers.Number):
                distance += dataid.get(key)
            elif isinstance(dataid.get(key), tuple):
                distance += len(dataid.get(key))
        return distance

    @staticmethod
    def _add_distance_from_query(dataid_val, requested_val, distance):
        try:
            distance += dataid_val.distance(requested_val)
        except AttributeError:
            if not isinstance(requested_val, list):
                requested_val = [requested_val]
            if dataid_val not in requested_val:
                distance = np.inf
            elif isinstance(dataid_val, numbers.Number):
                # so as to get the highest resolution first
                # FIXME: this ought to be clarified, not sure that
                # higher resolution is preferable is all cases.
                # Moreover this might break with other numerical
                # values.
                distance += dataid_val
        return distance

    def create_less_modified_query(self):
        """Create a query with one less modifier."""
        new_dict = self.to_dict()
        new_dict["modifiers"] = tuple(new_dict["modifiers"][:-1])
        return DataQuery.from_dict(new_dict)

    def is_modified(self):
        """Check if this is modified."""
        return bool(self._dict.get("modifiers"))


def create_filtered_query(dataset_key, filter_query):
    """Create a DataQuery matching *dataset_key* and *filter_query*.

    If a property is specified in both *dataset_key* and *filter_query*, the former
    has priority.

    """
    ds_dict = _create_id_dict_from_any_key(dataset_key)
    _update_dict_with_filter_query(ds_dict, filter_query)

    return DataQuery.from_dict(ds_dict)


def _update_dict_with_filter_query(ds_dict, filter_query):
    if filter_query is not None:
        for key, value in filter_query.items():
            if value != "*":
                ds_dict.setdefault(key, value)


def _create_id_dict_from_any_key(dataset_key):
    try:
        ds_dict = dataset_key.to_dict()
    except AttributeError:
        if isinstance(dataset_key, str):
            ds_dict = {"name": dataset_key}
        elif isinstance(dataset_key, numbers.Number):
            ds_dict = {"wavelength": dataset_key}
        else:
            raise TypeError("Don't know how to interpret a dataset_key of type {}".format(type(dataset_key)))
    return ds_dict



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for compositors using lookup tables."""

import unittest

import dask.array as da
import numpy as np
import xarray as xr


class TestColormapCompositor(unittest.TestCase):
    """Test the ColormapCompositor."""

    def setUp(self):
        """Set up the test case."""
        from satpy.composites.lookup import ColormapCompositor
        self.colormap_compositor = ColormapCompositor("test_cmap_compositor")

    def test_build_colormap_with_int_data_and_without_meanings(self):
        """Test colormap building."""
        palette = np.array([[0, 0, 0], [127, 127, 127], [255, 255, 255]])
        colormap, squeezed_palette = self.colormap_compositor.build_colormap(palette, np.uint8, {})
        assert np.allclose(colormap.values, [0, 1])
        assert np.allclose(squeezed_palette, palette / 255.0)

    def test_build_colormap_with_int_data_and_with_meanings(self):
        """Test colormap building."""
        palette = xr.DataArray(np.array([[0, 0, 0], [127, 127, 127], [255, 255, 255]]),
                               dims=["value", "band"])
        palette.attrs["palette_meanings"] = [2, 3, 4]
        colormap, squeezed_palette = self.colormap_compositor.build_colormap(palette, np.uint8, {})
        assert np.allclose(colormap.values, [2, 3, 4])
        assert np.allclose(squeezed_palette, palette / 255.0)


class TestPaletteCompositor(unittest.TestCase):
    """Test the PaletteCompositor."""

    def test_call(self):
        """Test palette compositing."""
        from satpy.composites.lookup import PaletteCompositor
        cmap_comp = PaletteCompositor("test_cmap_compositor")
        palette = xr.DataArray(np.array([[0, 0, 0], [127, 127, 127], [255, 255, 255]]),
                               dims=["value", "band"])
        palette.attrs["palette_meanings"] = [2, 3, 4]

        data = xr.DataArray(da.from_array(np.array([[4, 3, 2], [2, 3, 4]], dtype=np.uint8)), dims=["y", "x"])
        res = cmap_comp([data, palette])
        exp = np.array([[[1., 0.498039, 0.],
                         [0., 0.498039, 1.]],
                        [[1., 0.498039, 0.],
                         [0., 0.498039, 1.]],
                        [[1., 0.498039, 0.],
                         [0., 0.498039, 1.]]])
        assert np.allclose(res, exp)


class TestColorizeCompositor(unittest.TestCase):
    """Test the ColorizeCompositor."""

    def test_colorize_no_fill(self):
        """Test colorizing."""
        from satpy.composites.lookup import ColorizeCompositor
        colormap_composite = ColorizeCompositor("test_color_compositor")
        palette = xr.DataArray(np.array([[0, 0, 0], [127, 127, 127], [255, 255, 255]]),
                               dims=["value", "band"])
        palette.attrs["palette_meanings"] = [2, 3, 4]

        data = xr.DataArray(np.array([[4, 3, 2],
                                      [2, 3, 4]],
                                     dtype=np.uint8),
                            dims=["y", "x"])
        res = colormap_composite([data, palette])
        exp = np.array([[[1., 0.498039, 0.],
                         [0., 0.498039, 1.]],
                        [[1., 0.498039, 0.],
                         [0., 0.498039, 1.]],
                        [[1., 0.498039, 0.],
                         [0., 0.498039, 1.]]])
        assert np.allclose(res, exp, atol=0.0001)

    def test_colorize_with_interpolation(self):
        """Test colorizing with interpolation."""
        from satpy.composites.lookup import ColorizeCompositor
        colormap_composite = ColorizeCompositor("test_color_compositor")
        palette = xr.DataArray(np.array([[0, 0, 0], [127, 127, 127], [255, 255, 255]]),
                               dims=["value", "band"])
        palette.attrs["palette_meanings"] = [2, 3, 4]

        data = xr.DataArray(da.from_array(np.array([[4, 3, 2.5],
                                                    [2, 3.2, 4]])),
                            dims=["y", "x"],
                            attrs={"valid_range": np.array([2, 4])})
        res = colormap_composite([data, palette])
        exp = np.array([[[1.0, 0.498039, 0.246575],
                         [0., 0.59309977, 1.0]],
                        [[1.0, 0.49803924, 0.24657543],
                         [0., 0.59309983, 1.0]],
                        [[1.0, 0.4980392, 0.24657541],
                         [0., 0.59309978, 1.0]]])
        np.testing.assert_allclose(res, exp, atol=1e-4)


class TestCategoricalDataCompositor(unittest.TestCase):
    """Test composiotor for recategorization of categorical data."""

    def setUp(self):
        """Create test data."""
        attrs = {"name": "foo"}
        data = xr.DataArray(da.from_array([[2., 1.], [3., 0.]]), attrs=attrs,
                            dims=("y", "x"), coords={"y": [0, 1], "x": [0, 1]})

        self.data = data

    def test_basic_recategorization(self):
        """Test general functionality of compositor incl. attributes."""
        from satpy.composites.lookup import CategoricalDataCompositor
        lut = [np.nan, 0, 1, 1]
        name = "bar"
        comp = CategoricalDataCompositor(name=name, lut=lut)
        res = comp([self.data])
        res = res.compute()
        expected = np.array([[1., 0.], [1., np.nan]])
        np.testing.assert_equal(res.values, expected)
        np.testing.assert_equal(res.attrs["name"], name)
        np.testing.assert_equal(res.attrs["composite_lut"], lut)

    def test_too_many_datasets(self):
        """Test that ValueError is raised if more than one dataset is provided."""
        from satpy.composites.lookup import CategoricalDataCompositor
        lut = [np.nan, 0, 1, 1]
        comp = CategoricalDataCompositor(name="foo", lut=lut)
        np.testing.assert_raises(ValueError, comp, [self.data, self.data])
